"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""
import random
from typing import List, Dict, Callable

import numpy as np
from apricopt.solving.blackbox.BlackBox import BlackBox


class FunctionBlackBox(BlackBox):

    def __init__(self, dimension: int, function: Callable,
                 lower_bounds=None, upper_bounds=None):
        """

        :param function: A multivariate real function that takes a list of float and returns a float
        :param lower_bounds: A list of float such that the i-th element is the lower bound of the i-th variable
        :param upper_bounds: A list of float such that the i-th element is the upper bound of the i-th variable
        """
        self.d: int = dimension

        if lower_bounds is None:
            self.lower_bounds: List[float] = [float("-Inf") for _ in range(self.d)]
        else:
            if len(lower_bounds) != self.d:
                raise ValueError("The length of the lower bounds list must be equal to the dimension of the black box")
            self.lower_bounds = lower_bounds

        if upper_bounds is None:
            self.upper_bounds: List[float] = [float("Inf") for _ in range(self.d)]
        else:
            if len(lower_bounds) != self.d:
                raise ValueError("The length of the upprt bounds list must be equal to the dimension of the black box")
            self.upper_bounds = upper_bounds

        if len(self.lower_bounds) != len(self.upper_bounds):
            raise ValueError("The length of the lower bounds list must be equal to the length of the upper bounds list")

        for i in range(self.d):
            if self.lower_bounds[i] >= self.upper_bounds[i]:
                raise ValueError(f"The lower bound of the variable with index {i} is larger than the upper bound")

        self.function = function
        self._total_evaluations: int = 0
        self.parameters_ids: List[str] = []


    def is_input_valid(self, parameters: Dict[str, float]) -> bool:
        """

        :param parameters:
        :return:
        """
        try:
            if len(parameters) != self.d:
                return False
            low_ok: bool = all((parameters[x] >= self.lower_bounds[int(x)] for x in parameters.keys()))
            up_ok: bool = all((parameters[x] <= self.upper_bounds[int(x)] for x in parameters.keys()))
            return low_ok and up_ok
        except Exception as e:
            raise e

    def is_input_valid_array(self, parameters_array) -> bool:
        """

        :param parameters_array:
        :return:
        """
        param_ids = self.get_optimization_parameters_ids()
        parameters = {param_ids[i]: parameters_array[i] for i in range(len(parameters_array))}
        try:
            if len(parameters) != self.d:
                return False
            low_ok: bool = all((parameters[x] >= self.lower_bounds[int(x)] for x in parameters.keys()))
            up_ok: bool = all((parameters[x] <= self.upper_bounds[int(x)] for x in parameters.keys()))
            return low_ok and up_ok
        except Exception as e:
            raise e

    def evaluate(self, parameters: Dict[str, float], check_input=True) -> Dict[str, float]:
        """

        :param check_input:
        :param parameters:
        :return:
        """
        if check_input and not self.is_input_valid(parameters):
            raise ValueError("The provided input is not valid.")

        x = []
        for i in range(self.d):
            x.append(parameters[str(i)])

        self._total_evaluations += 1
        return {self.get_objective_id(): self.function(x)}

    def evaluate_np_array(self, parameters: np.array, check_input=False) -> Dict[str, float]:
        """

        :param check_input:
        :param parameters:
        :return:
        """
        if check_input and not self.is_input_valid_array(parameters):
            raise ValueError("The provided input is not valid.")
        self._total_evaluations += 1
        return {self.get_objective_id(): self.function(parameters)}

    def evaluate_objective_np_array(self, parameters: np.array, check_input=False) -> float:
        # param_ids = self.get_optimization_parameters_ids()
        # params = {param_ids[i]: parameters[i] for i in range(len(parameters))}
        if check_input and not self.is_input_valid_array(parameters):
            raise ValueError("The provided input is not valid.")
        self._total_evaluations += 1
        return self.function(parameters)
        # return self.evaluate(params, check_input)[self.get_objective_id()]

    def evaluate_objective_multiple_inputs_np_array(self, parameters_list: np.ndarray, check_input=False) \
            -> List[float]:
        results: List[float] = [0] * len(parameters_list)
        for i in range(len(parameters_list)):
            results[i] = self.evaluate_objective_np_array(parameters_list[i].tolist(), check_input)

        # results = self.function(parameters_list)
        return results

    def get_total_evaluations(self):
        return self._total_evaluations

    def get_optimization_parameters_number(self) -> int:
        return self.d

    def get_optimization_parameters_ids(self) -> List[str]:
        if not self.parameters_ids:
            self.parameters_ids = [str(i) for i in range(self.d)]
        return self.parameters_ids

    def get_optimization_parameter_lower_bound(self, param_id) -> float:
        return self.lower_bounds[int(param_id)]

    def get_optimization_parameter_upper_bound(self, param_id) -> float:
        return self.upper_bounds[int(param_id)]

    def get_optimization_parameters_lower_bounds_nparray(self) -> np.array:
        return np.array([self.get_optimization_parameter_lower_bound(p_id)
                         for p_id in self.get_optimization_parameters_ids()])

    def get_optimization_parameters_upper_bounds_nparray(self) -> np.array:
        return np.array([self.get_optimization_parameter_upper_bound(p_id)
                         for p_id in self.get_optimization_parameters_ids()])

    def get_optimization_parameter_initial_value(self, param_id) -> float:
        return random_in_range(self.get_optimization_parameter_lower_bound(param_id),
                               self.get_optimization_parameter_upper_bound(param_id))

    def get_random_optimization_parameters_initial_values(self) -> Dict[str, float]:
        return {param_id: random_in_range(self.get_optimization_parameter_lower_bound(param_id),
                                          self.get_optimization_parameter_upper_bound(param_id))
                for param_id in self.get_optimization_parameters_ids()}

    def get_random_initial_values_np_array(self) -> np.array:
        return np.array([random_in_range(self.get_optimization_parameter_lower_bound(param_id),
                                         self.get_optimization_parameter_upper_bound(param_id))
                         for param_id in self.get_optimization_parameters_ids()])

    def set_optimization_parameters_initial_values(self, param_values: Dict[str, float]) -> None:
        pass

    def get_optimization_parameter_granularity(self, param_id) -> float:
        return 0

    def get_extreme_barrier_constraints_number(self) -> int:
        return 0

    def get_progressive_barrier_constraints_number(self) -> int:
        return 0

    def get_extreme_barrier_constraints_ids(self) -> List[str]:
        return []

    def get_progressive_barrier_constraints_ids(self) -> List[str]:
        return []

    def get_objective_id(self) -> str:
        return "result"

    def get_objective_upper_bound(self) -> float:
        return float('Inf')

    @staticmethod
    def get_raisable_exception_type():
        return ValueError

    def finalize(self) -> None:
        pass


def random_in_range(lower_bound: float, upper_bound: float) -> float:
    """
    Returns a pseudo-random number in the given interval
    Restituisce un numero reale (pseudo-)casuale nell'intervallo dato
    :param lower_bound: the lower bound of the interval
    :param upper_bound: the upper bound of the interval
    :return: a random flaot c such that lower_bound <= c <= upper_bound
    """
    lb = 0 if lower_bound == float("-Inf") else lower_bound
    ub = 1 if upper_bound == float("Inf") else upper_bound
    return lb + random.random() * (ub - lb)



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

import COPASI
from apricopt.solving.whitebox.COPASI.COPASIOptimisationMethod import COPASIOptimisationMethod
from apricopt.solving.whitebox.COPASI.COPASISolverParameter import COPASISolverParameter
from typing import List


class COPASIEvolutionaryProgramming(COPASIOptimisationMethod):
    NUMBER_OF_GENERATIONS = "Number of Generations"
    POPULATION_SIZE = "Population Size"
    SEED = "Seed"
    RANDOM_NUMBER_GENERATOR = "Random Number Generator"
    STOP_AFTER_N_STALLED_GENERATIONS = "Stop after # Stalled Generations"

    def __init__(self):
        super(COPASIEvolutionaryProgramming, self).__init__()
        self.solver_parameters = [
            COPASISolverParameter(COPASIEvolutionaryProgramming.NUMBER_OF_GENERATIONS, 200),
            COPASISolverParameter(COPASIEvolutionaryProgramming.POPULATION_SIZE, 20)
        ]

    def set_parameters_configuration(self, opt_task: COPASI.COptTask,
                                     solver_parameters: List[COPASISolverParameter] = None):
        solver_params = self.solver_parameters if solver_parameters is None else solver_parameters
        self.set_optimisation_task(opt_task)
        self.opt_task.setMethodType(COPASI.CTaskEnum.Method_EvolutionaryProgram)
        self.set_and_check_solver_parameters(solver_params)




"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

from apricopt.model.Observable import Observable
from apricopt.model.ObservableFunction import *

obs_function: Callable = MaxDistanceFromBounds
o = Observable('obs_0', "test", ["x + y", "z", "70"], obs_function)

assignment = {'time': [0, 1, 2, 3, 4, 5, 6, 7],
              "x": [0, 1, 2, 3, 4, 5, 6, 7],
              "y": [54, 24, 65, 67, 23, 45, 76, 12],
              "z": [5, 5, 5, 5, 5, 5, 5, 5]}

print(o.evaluate(assignment))



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2022 Sapienza University of Rome, Marco Esposito, Leonardo Picchiami.
"""

from typing import Dict, List, Type
import COPASI
import sys

from apricopt.model.Model import Model
from apricopt.simulation.COPASI.COPASIModelInstance import COPASIModelInstance
from apricopt.simulation.SimulationEngine import SimulationEngine


class COPASIEngine(SimulationEngine):

    def __init__(self):
        super().__init__()

    def load_model(self, model_filename: str, **exec_info) -> COPASIModelInstance:
        data_model = COPASI.CRootContainer.addDatamodel()
        assert (isinstance(data_model, COPASI.CDataModel))
        try:
            res = data_model.importSBML(model_filename)

            if not res:
                sys.stderr.write("There was an error loading the SBML model.\n")
                exit(1)
            return COPASIModelInstance(data_model)
        except COPASI.CCopasiException:
            sys.stderr.write("There was an error loading the SBML model.\n")
            exit(2)

    def simulate_trajectory(self, model: Model, horizon: float) -> Dict[str, List[float]]:
        if not isinstance(model.instance, COPASIModelInstance):
            raise TypeError("The object in the field 'instance' of a 'Model' object passed to "
                            "'COPASIEngine::simulate_trajectory_and_set' method must have type 'COPASIModelInstance'.")
        time_series: COPASI.CTimeSeries = self._run_simulation_task(model, horizon)
        return self._get_simulation_output(model, time_series)

    def simulate_trajectory_and_set(self, model: Model, horizon: float, exclude=None) -> Dict[str, List[float]]:
        if not isinstance(model.instance, COPASIModelInstance):
            raise TypeError("The object in the field 'instance' of a 'Model' object passed to "
                            "'COPASIEngine::simulate_trajectory_and_set' method must have type 'COPASIModelInstance'.")
        if exclude is None:
            exclude = []
        time_series: COPASI.CTimeSeries = self._run_simulation_task(model, horizon)
        model_instance: COPASIModelInstance = model.instance
        data_model: COPASI.CDataModel = model_instance.model_obj
        changed_params: Dict[str, float] = dict()
        changed_metabolites: Dict[str, float] = dict()
        changed_compartments: Dict[str, float] = dict()

        for out_i in range(1, time_series.getNumVariables()):
            output_id = time_series.getSBMLId(out_i, data_model)
            if output_id not in exclude:
                if output_id in model_instance.get_metabolites_ids():
                    changed_metabolites[output_id] = time_series.getConcentrationData(
                        time_series.getRecordedSteps() - 1,
                        out_i)
                elif output_id in model_instance.get_values_ids():
                    changed_params[output_id] = time_series.getData(time_series.getRecordedSteps() - 1, out_i)
                elif output_id in model_instance.get_compartments_ids():
                    changed_compartments[output_id] = time_series.getData(time_series.getRecordedSteps() - 1, out_i)
                elif output_id == data_model.getModel().getObjectName():
                    continue
                else:
                    # sys.stderr.write(f"There was an error collecting the output for variable with id {output_id}.\n")
                    continue

        model_instance.set_species_initial_values(changed_metabolites)
        model_instance.set_parameters(changed_params)
        model_instance.set_compartments_initial_sizes(changed_compartments)

        return self._get_simulation_output(model, time_series)

    def simulate_trajectory_and_get_state(self, model: Model, horizon: float, exclude=None) -> \
            Dict[str, Dict[str, float]]:
        if not isinstance(model.instance, COPASIModelInstance):
            raise TypeError(
                "The object in the field 'instance' of a 'Model' object passed to "
                "'COPASIEngine::simulate_trajectory_and_get_state' method must have type 'COPASIModelInstance'.")

        if exclude is None:
            exclude = []
        time_series: COPASI.CTimeSeries = self._run_simulation_task(model, horizon)
        model_instance: COPASIModelInstance = model.instance
        data_model: COPASI.CDataModel = model_instance.model_obj
        changed_params: Dict[str, float] = dict()
        changed_metabolites: Dict[str, float] = dict()
        changed_compartments: Dict[str, float] = dict()

        for out_i in range(1, time_series.getNumVariables()):
            output_id = time_series.getSBMLId(out_i, data_model)
            if output_id not in exclude:
                if output_id in model_instance.get_metabolites_ids():
                    changed_metabolites[output_id] = time_series.getConcentrationData(
                        time_series.getRecordedSteps() - 1,
                        out_i)
                elif output_id in model_instance.get_values_ids():
                    changed_params[output_id] = time_series.getData(time_series.getRecordedSteps() - 1, out_i)
                elif output_id in model_instance.get_compartments_ids():
                    changed_compartments[output_id] = time_series.getData(time_series.getRecordedSteps() - 1, out_i)
                elif output_id == data_model.getModel().getObjectName():
                    continue
                else:
                    # sys.stderr.write(f"There was an error collecting the output for variable with id {output_id}.\n")
                    continue

        result = dict()
        result['changed_metabolites'] = changed_metabolites
        result['changed_params'] = changed_params
        result['changed_compartments'] = changed_compartments

        return result

    def restore_state(self, model: Model, changed_values: dict) -> None:
        model_instance = model.instance
        if not isinstance(model_instance, COPASIModelInstance):
            raise TypeError(
                "The object in the field 'instance' of a 'Model' object passed to "
                "'COPASIEngine::simulate_trajectory_and_get_state' method must have type 'COPASIModelInstance'.")
        model_instance.set_species_initial_values(changed_values['changed_metabolites'])
        model_instance.set_parameters(changed_values['changed_params'])
        model_instance.set_compartments_initial_sizes(changed_values['changed_compartments'])

    def _get_simulation_output(self, model: Model, time_series: COPASI.CTimeSeries) -> Dict[str, List[float]]:
        if not isinstance(model.instance, COPASIModelInstance):
            raise TypeError(
                "The object in the field 'instance' of a 'Model' object passed to "
                "'COPASIEngine::simulate_trajectory_and_set' method must have type 'COPASIModelInstance'.")
        sim_output: Dict[str, List[float]] = dict()
        sim_output['time'] = [time_series.getData(s, 0) for s in range(time_series.getRecordedSteps())]

        for i in range(time_series.getNumVariables()):
            output_id = time_series.getSBMLId(i, model.instance.model_obj)
            if output_id in model.observed_outputs or not model.observed_outputs:
                # model_instance: COPASIModelInstance = model.instance
                # if output_id in model_instance.get_values_ids() or output_id in model_instance.get_compartments_ids():
                # if the observable was the concentration of a species we should use time_series.getConcentrationData(s,i)
                sim_output[output_id] = [time_series.getData(s, i)
                                         for s in range(time_series.getRecordedSteps())]
        return sim_output

    def _run_simulation_task(self, model: Model, horizon: float) -> COPASI.CTimeSeries:
        if not isinstance(model.instance, COPASIModelInstance):
            raise TypeError("The object in the field 'instance' of a 'Model' object passed to "
                            "'COPASIEngine::simulate_trajectory_and_set' method must have type 'COPASIModelInstance'.")
        model_instance: COPASIModelInstance = model.instance
        data_model: COPASI.CDataModel = model_instance.model_obj
        simulation_task = data_model.getTask("Time-Course")
        assert simulation_task is not None
        copasi_model = data_model.getModel()

        copasi_model.setInitialTime(0.0)
        simulation_task.setMethodType(COPASI.CTaskEnum.Method_deterministic)
        simulation_task.setScheduled(True)

        model_instance.set_simulation_duration_with_step_size(horizon, model.time_step)

        # TODO better error handling
        '''
        sim_ok = True
        try:
            sim_ok = simulation_task.process(True)
        except:
            sys.stderr.write("Error. Running the time course simulation failed.\n")
            sys.stderr.write("{0}\n".format(simulation_task.getProcessWarning()))
            sys.stderr.write("{0}\n".format(simulation_task.getProcessError())) 
            # Check if there are additional error messages
            if COPASI.CCopasiMessage.size() > 0:
                # print the messages in chronological order
                sys.stderr.write("{0}\n".format(COPASI.CCopasiMessage.getAllMessageText(True)))
            #exit(-1)
        if not sim_ok:
            sys.stderr.write("An error occured while running the time course simulation.\n")
            sys.stderr.write("{0}\n".format(simulation_task.getProcessWarning()))
            sys.stderr.write("{0}\n".format(simulation_task.getProcessError()))
            # Check if there are additional error messages
            if COPASI.CCopasiMessage.size() > 0:
                # Print the messages in chronological order
                sys.stderr.write("{0}\n".format(COPASI.CCopasiMessage.getAllMessageText(True)))
            #exit(-1)
        '''

        sim_ok = simulation_task.process(True)
        time_series: COPASI.CTimeSeries = simulation_task.getTimeSeries()
        return time_series

    def model_instance_class(self) -> Type[COPASIModelInstance]:
        return COPASIModelInstance



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2022 Sapienza University of Rome, Marco Esposito, Leonardo Picchiami.
"""

from typing import Type, Dict, List

from apricopt.model.Model import Model
from apricopt.model.ModelInstance import ModelInstance
from apricopt.simulation.SimulationEngine import SimulationEngine

import roadrunner
from roadrunner import RoadRunner
import numpy as np

from apricopt.simulation.roadrunner.RoadRunnerModelInstance import RoadRunnerModelInstance


class RoadRunnerEngine(SimulationEngine):

    def __init__(self):
        super().__init__()

    def load_model(self, model_filename: str, **exec_info) -> ModelInstance:
        model_obj: RoadRunner = roadrunner.RoadRunner(model_filename)
        return RoadRunnerModelInstance(model_obj)

    def simulate_trajectory(self, model: Model, horizon: float) -> Dict[str, List[float]]:
        sim_output = self._run_simulation_task(model, horizon)
        model_obj: RoadRunner = model.instance.model_obj
        model_obj.resetAll()
        return sim_output

    def simulate_trajectory_and_set(self, model: Model, horizon: float, exclude=None) -> Dict[str, List[float]]:
        model_obj: RoadRunner = model.instance.model_obj
        observed_outputs = model_obj.model.getFloatingSpeciesIds() + \
                           model_obj.model.getGlobalParameterIds() + \
                           model_obj.model.getCompartmentIds()

        sim_output = self._run_simulation_task(model, horizon, observed_outputs)
        exclude_from_set = self._exclude_from_simulate_and_set(model, exclude)

        changed_values: Dict[str, List[float]] = dict()
        for out_id, traj in sim_output.items():
            if out_id == 'time':
                changed_values['time'] = traj
                continue

            if out_id not in exclude_from_set:
                changed_values[out_id] = traj
                model_obj.model[f'init({out_id})'] = traj[-1]

        model_obj.resetAll()
        return changed_values

    def simulate_trajectory_and_get_state(self, model: Model, horizon: float, exclude=None) -> Dict[str, float]:
        model_obj: RoadRunner = model.instance.model_obj
        observed_outputs = model_obj.model.getFloatingSpeciesIds() + \
                           model_obj.model.getGlobalParameterIds() + \
                           model_obj.model.getCompartmentIds()

        sim_output = self._run_simulation_task(model, horizon, observed_outputs)
        exclude_from_set = self._exclude_from_simulate_and_set(model, exclude)
        changed_values: Dict[str, float] = dict()
        for out_id, traj in sim_output.items():
            if out_id == 'time':
                continue

            if out_id not in exclude_from_set:
                changed_values[out_id] = traj[-1]

        model_obj.resetAll()
        return changed_values

    def restore_state(self, model: Model, changed_values: Dict[str, float]) -> None:
        model_obj: RoadRunner = model.instance.model_obj
        for var_id, state_value in changed_values.items():
            model_obj[f'init({var_id})'] = state_value
        model_obj.resetAll()

    def model_instance_class(self) -> Type[ModelInstance]:
        return RoadRunnerModelInstance

    def _exclude_from_simulate_and_set(self, model: Model, exclude: List[str] = None) -> List[str]:
        model_obj = model.instance.model_obj

        init_exclude: List[str] = []
        if exclude:
            init_exclude = list(exclude)

        for init_ass in model_obj.getInitialAssignmentIds():
            if init_ass not in init_exclude:
                init_exclude.append(init_ass)

        for ass in model_obj.getAssignmentRuleIds():
            if ass not in init_exclude:
                init_exclude.append(ass)
        return init_exclude

    def _get_simulation_output(self, model: Model, time_series: np.ndarray, observed) -> Dict[str, List[float]]:
        if not isinstance(model.instance, RoadRunnerModelInstance):
            raise TypeError("The object in the field 'instance' of a 'Model' object passed to "
                            "'RoadRunnerEngine::simulate_trajectory' method must have type 'RoadRunnerModelInstance'.")
        sim_output: Dict[str, List[float]] = dict()
        sim_output['time'] = time_series[:, 0]
        for i in range(1, len(observed)):
            sim_output[observed[i]] = list(time_series[:, i])

        return sim_output

    def _run_simulation_task(self, model: Model, horizon: float, observed_outputs=None) -> np.ndarray:
        if not isinstance(model.instance, RoadRunnerModelInstance):
            raise TypeError("The object in the field 'instance' of a 'Model' object passed to "
                            "'RoadRunnerEngine::simulate_trajectory' method must have type 'RoadRunnerModelInstance'.")
        model_obj: RoadRunner = model.instance.model_obj
        if model.instance.time_step == -1:
            raise RuntimeError("The time step must be set before running the simulation.")
        observed: List[str]

        if observed_outputs:
            observed = observed_outputs
        elif model.observed_outputs:
            observed = model.observed_outputs
        else:
            observed = model_obj.model.getFloatingSpeciesIds()

        time_observation = None
        for observation in observed:
            if observation.lower() == 'time':
                time_observation = observation

        if time_observation:
            observed.remove(time_observation)

        observed = ['time'] + observed
        model_obj.timeCourseSelections = observed
        trajectory = model_obj.simulate(0, horizon, steps=int(horizon / model.instance.time_step))
        return self._get_simulation_output(model, trajectory, observed)



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2022 Sapienza University of Rome, Marco Esposito, Leonardo Picchiami.
"""


class Parameter:

    def __init__(self, param_id: str, name: str, lower_bound: float = float("-inf"),
                 upper_bound: float = float("inf"), nominal_value: float = float("-inf"),
                 distribution: str = 'uniform', mu: float = 0, sigma: float = 0,
                 granularity: float = 0, optimize: bool = True):
        self.id = param_id
        self.name = name
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.nominal_value = nominal_value
        self.distribution = distribution
        self.mu = mu
        self.sigma = sigma
        self.granularity = granularity
        self.optimize = optimize



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

import COPASI

from typing import *

from apricopt.model.ModelInstance import ModelInstance


class COPASIModelInstance(ModelInstance):

    def __init__(self, model_obj: COPASI.CDataModel):
        super().__init__(model_obj)
        self.cached_metabolites_ids = None
        self.cached_values_ids = None
        self.cached_compartments_ids = None
        self.cached_value_objects = dict()
        self.cached_metabolite_objects = dict() # added merge

    def __delete__(self, instance):
        model_obj: COPASI.CDataModel = self.model_obj
        model_obj.destruct()

    def set_parameters(self, params_values: Dict[str, float]) -> None:
        copasi_model = self.model_obj.getModel()

        changed_objects = COPASI.ObjectStdVector()
        assert changed_objects is not None
        for param_id, value in params_values.items():
            param_obj = self.getModelValueBySBMLId(copasi_model, param_id)
            param_obj.setInitialValue(value)
            ivr = param_obj.getInitialValueReference()
            changed_objects.push_back(ivr)
        copasi_model.updateInitialValues(changed_objects)

    def getModelValueBySBMLId(self, copasi_model: COPASI.CModel, value_id: str) -> COPASI.CModelValue:
        if not value_id in self.cached_value_objects:
            found = False
            for value in copasi_model.getModelValues():
                if value.getSBMLId() == value_id:
                    self.cached_value_objects[value_id] = value
                    found = True
                    break
            if not found:
                raise ValueError(f"Could not find model parameter with id {value_id}")

        return self.cached_value_objects[value_id]

    def getModelValueBySBMLIdIfPresent(self, copasi_model: COPASI.CModel, value_id: str) -> Union[COPASI.CModelValue, None]:
        if not value_id in self.cached_value_objects:
            for value in copasi_model.getModelValues():
                if value.getSBMLId() == value_id:
                    self.cached_value_objects[value_id] = value
                    return self.cached_value_objects[value_id]
            return None
        return self.cached_value_objects[value_id]

    def getMetaboliteBySBMLIdIfPresent(self, copasi_model: COPASI.CModel, value_id: str) -> Union[COPASI.CMetab, None]:
        if not value_id in self.cached_metabolite_objects:
            for value in copasi_model.getMetabolites():
                if value.getSBMLId() == value_id:
                    self.cached_metabolite_objects[value_id] = value
                    return self.cached_metabolite_objects[value_id]
            return None
        return self.cached_metabolite_objects[value_id]

    def set_compartments_initial_sizes(self, compartments_values: Dict[str, float]) -> None:
        copasi_model: COPASI.CModel = self.model_obj.getModel()
        changed_objects = COPASI.ObjectStdVector()
        assert changed_objects is not None
        for compartment_id, compartment_size in compartments_values.items():
            # compartment_obj: COPASI.CCompartment = self.get_compartment_by_id(compartment_id)
            compartment_obj: COPASI.CCompartment = self.model_obj.getModel().getCompartment(compartment_id)
            compartment_obj.setInitialValue(compartment_size)
            ivr = compartment_obj.getInitialValueReference()
            changed_objects.push_back(ivr)

            # self.model_obj.getModel().getCompartment(compartment_id).setInitialValue(compartment_size)

            copasi_model.updateInitialValues(changed_objects)

    def set_species_initial_values(self, values: Dict[str, float]) -> None:
        changedObjects = COPASI.ObjectStdVector()
        for met_id, met_value in values.items():
            found = False
            for metabolite_obj in self.model_obj.getModel().getMetabolites():
                if metabolite_obj.getSBMLId() == met_id:
                    found = True
                    metabolite_obj.setInitialValue(metabolite_obj.getValue())
                    icr = metabolite_obj.getInitialValueReference()
                    changedObjects.push_back(icr)
            if not found:
                raise Exception(f"ERROR: Metabolite with id {met_id} not found in the model.")
        self.model_obj.getModel().updateInitialValues(changedObjects)

    def set_species_initial_values_OLD(self, values: Dict[str, float]) -> None:
        for met_id, met_value in values.items():
            self.set_species_initial_value(met_id, met_value)

    def set_species_initial_value(self, metabolite_id: str, metabolite_value: float) -> None:
        '''metabolite_obj: COPASI.CMetab = self.get_metabolite_by_id(metabolite_id)
        metabolite_obj.setInitialConcentration(metabolite_value)
        self.model_obj.getModel().updateInitialValues(metabolite_obj.getInitialConcentrationReference())'''

        for metabolite_obj in self.model_obj.getModel().getMetabolites():
            if metabolite_obj.getSBMLId() == metabolite_id:
                # metabolite_obj.setInitialConcentration(metabolite_value)
                metabolite_obj.setInitialValue(metabolite_obj.getValue())
                # metabolite_obj.setInitialConcentration(metabolite_obj.getConcentration())
                changedObjects = COPASI.ObjectStdVector()
                # icr = metabolite_obj.getInitialConcentrationReference()
                icr = metabolite_obj.getInitialValueReference()
                changedObjects.push_back(icr)
                self.model_obj.getModel().updateInitialValues(changedObjects)
                return
        raise Exception(f"Metabolite with id {metabolite_id} not found in the model")
        # metabolite_obj: COPASI.CMetab = self.model_obj.getModel().getMetabolite(metabolite_id)

    def get_parameters_initial_values(self) -> Dict[str, float]:
        copasi_model: COPASI.CModel = self.model_obj.getModel()
        result: Dict[str, float] = dict()
        for param in copasi_model.getModelValues():
            result[param.getSBMLId()] = param.initial_value
        return result

    def get_compartment_initial_volumes(self) -> Dict[str, float]:
        copasi_model: COPASI.CModel = self.model_obj.getModel()
        result: Dict[str, float] = dict()
        for comp in copasi_model.getModelValues():
            result[comp.getSBMLId()] = comp.getInitialValue()
        return result

    def set_parameter(self, param_id: str, value: float) -> None:
        self.set_parameters({param_id: value})

    def set_simulation_configuration(self, abs_tol: float, rel_tol: float, step_size: float) -> None:
        simulation_task = self.model_obj.getTask("Time-Course")
        assert simulation_task is not None

        # Relative and absolute tolerance
        method = simulation_task.getMethod()
        absolute_tolerance = method.getParameter("Absolute Tolerance")
        assert absolute_tolerance is not None
        assert absolute_tolerance.getType() == COPASI.CCopasiParameter.Type_UDOUBLE
        absolute_tolerance.setValue(abs_tol)

        relative_tolerance = method.getParameter("Relative Tolerance")
        assert relative_tolerance is not None
        assert relative_tolerance.getType() == COPASI.CCopasiParameter.Type_UDOUBLE
        relative_tolerance.setValue(rel_tol)

        simulation_problem = simulation_task.getProblem()
        assert simulation_problem is not None

        simulation_problem.setModel(self.model_obj.getModel())
        simulation_problem.setStepSize(step_size)
        simulation_problem.setTimeSeriesRequested(True)

    def set_simulation_duration(self, duration: float) -> None:
        simulation_task = self.model_obj.getTask("Time-Course")
        assert simulation_task is not None
        simulation_problem: COPASI.CTrajectoryProblem = simulation_task.getProblem()
        assert simulation_problem is not None

        simulation_problem.setModel(self.model_obj.getModel())
        simulation_problem.setDuration(duration)

        simulation_problem.setStepNumber(int(duration / simulation_problem.getStepSize()))

    def set_simulation_duration_with_step_size(self, duration: float, step_size: float) -> None:
        simulation_task = self.model_obj.getTask("Time-Course")
        assert simulation_task is not None
        simulation_problem: COPASI.CTrajectoryProblem = simulation_task.getProblem()
        assert simulation_problem is not None

        simulation_problem.setModel(self.model_obj.getModel())
        simulation_problem.setDuration(duration)

        simulation_problem.setStepNumber(int(duration / step_size))

    def get_metabolites_ids(self) -> Set[str]:
        if not self.cached_metabolites_ids:
            self.cached_metabolites_ids: Set[str] = set()
            metabolites: COPASI.MetabVector = self.model_obj.getModel().getMetabolites()
            for metabolite in metabolites:
                self.cached_metabolites_ids.add(metabolite.getSBMLId())

        return self.cached_metabolites_ids

    def get_values_ids(self) -> Set[str]:
        if not self.cached_values_ids:
            self.cached_values_ids: Set[str] = set()
            model_values: COPASI.ModelValueVectorN = self.model_obj.getModel().getModelValues()
            for model_value in model_values:
                self.cached_values_ids.add(model_value.getSBMLId())
        return self.cached_values_ids

    def get_compartments_ids(self) -> Set[str]:
        if not self.cached_compartments_ids:
            self.cached_compartments_ids: Set[str] = set()
            compartments: COPASI.CompartmentVectorNS = self.model_obj.getModel().getCompartments()
            for compartment in compartments:
                self.cached_compartments_ids.add(compartment.getSBMLId())
        return self.cached_compartments_ids

    '''def get_metabolites_by_id(self) -> Dict[str, COPASI.CMetab]:
        if not self.cached_metabolites_by_id:
            self.cached_metabolites_by_id: Dict[str, COPASI.CMetab] = dict()
            metabolites: COPASI.MetabVector = self.model_obj.getModel().getMetabolites()
            for metabolite in metabolites:
                self.cached_metabolites_by_id[metabolite.getSBMLId()] = metabolite

        return self.cached_metabolites_by_id

    def get_metabolite_by_id(self, met_id:str) -> COPASI.CMetab:
        return self.get_metabolites_by_id()[met_id]

    def get_values_by_id(self) -> Dict[str, COPASI.CModelValue]:
        if not self.cached_values_by_id:
            self.cached_values_by_id: Dict[str, COPASI.CModelValue] = dict()
            model_values: COPASI.ModelValueVectorN = self.model_obj.getModel().getModelValues()
            for model_value in model_values:
                self.cached_values_by_id[model_value.getSBMLId()] = model_value
        return self.cached_values_by_id

    def get_value_by_id(self, value_id: str) -> COPASI.CModelValue:
        return self.get_values_by_id()[value_id]

    def get_compartments_by_id(self) -> Dict[str, COPASI.CCompartment]:
        if not self.cached_compartments_by_id:
            self.cached_compartments_by_id: Dict[str, COPASI.CCompartment] = dict()
            compartments: COPASI.CompartmentVectorNS = self.model_obj.getModel().getCompartments()
            for compartment in compartments:
                self.cached_compartments_by_id[compartment.getSBMLId()] = compartment
        return self.cached_compartments_by_id

    def get_compartment_by_id(self, compartment_id: str) -> COPASI.CCompartment:
        return self.get_compartments_by_id()[compartment_id]'''



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

from typing import Dict, List, Callable

from apricopt.model.Observable import Observable, compute_expressions_values_from_assignment


class FastObservable(Observable):

    def __init__(self, param_id: str, name: str, expression: List[str], function: Callable):
        super(FastObservable, self).__init__(param_id, name, expression, function)

    def evaluate(self, assignment: Dict[str, float]) -> float:
        expressions_values = compute_expressions_values_from_assignment(assignment, self.expressions, self.function)
        ret = self.function(*expressions_values)
        return ret



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

from abc import ABC, abstractmethod
from typing import Dict, Any

from apricopt.solving.blackbox.BlackBox import BlackBox


class BlackBoxSolver(ABC):

    def __init__(self):
        self.times = []
        self.objective_values = []
        self.admissible = []
        self.extreme_barrier_constraints_values = {}
        self.progressive_barrier_constraints_values = {}
        self.log = []
        self.start_time = -1

    @abstractmethod
    def solve(self, black_box: BlackBox,
              solver_params: Dict[str, Any]) -> (Dict[str, float], float, float, float, float):
        pass

    def initialize_storage(self, black_box: BlackBox) -> None:
        self.times = []
        self.objective_values = []
        self.extreme_barrier_constraints_values = {}
        self.progressive_barrier_constraints_values = {}
        for extreme_constraint_id in black_box.get_extreme_barrier_constraints_ids():
            self.extreme_barrier_constraints_values[extreme_constraint_id] = []
        for progressive_constraint_id in black_box.get_progressive_barrier_constraints_ids():
            self.progressive_barrier_constraints_values[progressive_constraint_id] = []
        self.log = []

    def set_start_time(self, start_time) -> None:
        self.start_time = start_time

    def add_extreme_barrier_constraint_value(self, constraint_id, value):
        self.extreme_barrier_constraints_values[constraint_id].append(value)

    def add_progressive_barrier_constraint_value(self, constraint_id, value):
        self.progressive_barrier_constraints_values[constraint_id].append(value)

    def add_objective_value(self, value):
        self.objective_values.append(value)



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020 Marco Esposito, Leonardo Picchiami.
"""

from typing import Dict
from apricopt.model.ModelInstance import ModelInstance


class MockUpModelInstance(ModelInstance):

    def __init__(self, model_obj):
        super().__init__(model_obj)

    def set_parameter(self, param_name: str, value: float) -> None:
        pass

    def set_parameters(self, params_values: Dict[str, float]) -> None:
        pass

    def get_parameters_initial_values(self) -> Dict[str, float]:
        return {"parameter": 0.5}

    def get_compartment_initial_volumes(self) -> Dict[str, float]:
        return dict()

    def set_simulation_configuration(self, abs_tol: float, rel_tol: float, step_size: float) -> None:
        pass

    def set_simulation_duration(self, duration: float) -> None:
        pass

    def set_simulation_duration_with_step_size(self, horizon, time_step):
        pass


"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

from typing import Dict
import random

from apricopt.solving.blackbox.BlackBox import BlackBox
from apricopt.solving.blackbox.BlackBoxSolver import BlackBoxSolver


class MockUpSolver(BlackBoxSolver):

    def __init__(self):
        super().__init__()

    def solve(self, black_box: BlackBox,
              solver_params: list) -> (Dict[str, float], float, float):
        min_obj = float("Inf")
        min_params = None
        for i in range(solver_params[0]):
            print(f"Iteration {i+1}")
            params = dict()
            for param_id in black_box.get_optimization_parameters_ids():
                params[param_id] = random.uniform(black_box.get_optimization_parameter_lower_bound(param_id),
                                                  black_box.get_optimization_parameter_upper_bound(param_id))
            print(f"Parameters: {params}")

            params_values_dict = params
            print(f"\nTrying with params: {params_values_dict}")
            result: Dict[str, float] = black_box.evaluate(params_values_dict)
            extreme_barrier_violated = False
            for extreme_barrier_id in black_box.get_extreme_barrier_constraints_ids():
                if result[extreme_barrier_id] > 0:
                    extreme_barrier_violated = True
                    print(f"Violated Extreme Barrier Constraint: {extreme_barrier_id} with value "
                          f"{result[extreme_barrier_id]}")
                    break

            if not extreme_barrier_violated:
                objective: float = result[black_box.get_objective_id()]
                print(f"Objective value: {objective}")
                if objective <= min_obj:
                    min_obj = objective
                    min_params = params

        return min_params, min_obj, solver_params[0], solver_params[0], solver_params[0]



from typing import List, Dict

from apricopt.model.Model import Model
from apricopt.model.Observable import Observable
from apricopt.model.Parameter import Parameter
from apricopt.simulation.MockUp.MockUpSimulationEngine import MockUpSimulationEngine
from apricopt.solving.blackbox.BlackBox import BlackBox


class MockUpBlackBox(BlackBox):

    def __init__(self, sim_engine: MockUpSimulationEngine):
        self.sim_engine = sim_engine
        self.model = Model(sim_engine, "", 0, 0, 1, observed_outputs=["output"])
        self.model.objective = Observable("output", "output", ["output"])
        parameter = Parameter("parameter", "parameter", lower_bound=0, upper_bound=1, nominal_value=0.5)
        self.model.set_parameter_space({parameter})
        self.horizon = 1

    def evaluate(self, parameters: Dict[str, float]) -> Dict[str, float]:
        return self.sim_engine.simulate(self.model, self.horizon)

    def get_optimization_parameters_number(self) -> int:
        return len(self.model.parameters)

    def get_optimization_parameters_ids(self) -> List[str]:
        ids = [p_id for p_id in self.model.parameters.keys()]
        ids.sort()
        return ids

    def get_optimization_parameter_lower_bound(self, param_id) -> float:
        return self.model.parameters[param_id].lower_bound

    def get_optimization_parameter_upper_bound(self, param_id) -> float:
        return self.model.parameters[param_id].upper_bound

    def get_optimization_parameter_initial_value(self, param_id) -> float:
        return self.model.parameters[param_id].nominal_value

    def set_optimization_parameters_initial_values(self, param_values: Dict[str, float]) -> None:
        for param_id, param_value in param_values.items():
            if self.get_optimization_parameter_lower_bound(param_id) <= \
                    param_value <= self.get_optimization_parameter_upper_bound(param_id):
                self.model.parameters[param_id].nominal_value = param_value
            else:
                raise ValueError(f"Value {param_value} for parameter {param_id} is outside the bounds")

    def get_optimization_parameter_granularity(self, param_id) -> float:
        return self.model.parameters[param_id].granularity

    def get_extreme_barrier_constraints_number(self) -> int:
        return 0

    def get_progressive_barrier_constraints_number(self) -> int:
        return 0

    def get_extreme_barrier_constraints_ids(self) -> List[str]:
        return []

    def get_progressive_barrier_constraints_ids(self) -> List[str]:
        return []

    def get_objective_id(self) -> str:
        return "output"

    def get_objective_upper_bound(self) -> float:
        return 1

    @staticmethod
    def get_raisable_exception_type():
        pass

    def finalize(self) -> None:
        pass



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

from typing import Dict

from apricopt.model.ModelInstance import ModelInstance
from roadrunner import RoadRunner, ExecutableModel



class RoadRunnerModelInstance(ModelInstance):

    def __init__(self, model_obj: RoadRunner):
        super().__init__(model_obj)
        self.time_step = -1
        self.horizon = -1

    def __delete__(self, instance):
        model_obj: RoadRunner = self.model_obj
        del model_obj

    def set_parameter(self, param_name: str, value: float) -> None:
        model_obj: RoadRunner = self.model_obj
        model_obj.model[f'init({param_name})'] = value
        model_obj.resetAll()

    def set_parameters(self, params_values: Dict[str, float]) -> None:
        model_obj: RoadRunner = self.model_obj
        for param_id, value in params_values.items():
            model_obj.model[f'init({param_id})'] = value
        model_obj.resetAll()

    def get_parameters_initial_values(self) -> Dict[str, float]:
        values: Dict[str, float] = dict()
        model_obj: RoadRunner = self.model_obj
        ex_model: ExecutableModel = model_obj.model
        params_ids = ex_model.getGlobalParameterIds()
        for i in range(len(params_ids)):
            values[params_ids[i]] = ex_model['init('+params_ids[i]+')']
        return values

    def get_compartment_initial_volumes(self) -> Dict[str, float]:
        values: Dict[str, float] = dict()
        model_obj: RoadRunner = self.model_obj
        ex_model: ExecutableModel = model_obj.model
        ids = ex_model.getCompartmentIds()
        init_volumes = ex_model.getCompartmentInitVolumes()
        for i in range(len(ids)):
            values[ids[i]] = init_volumes[i]
        return values

    def set_simulation_configuration(self, abs_tol: float, rel_tol: float, step_size: float) -> None:
        self.model_obj.integrator.stiff = True
        self.model_obj.integrator.initial_time_step = 0
        self.model_obj.integrator.absolute_tolerance = abs_tol
        self.model_obj.integrator.relative_tolerance = rel_tol
        self.time_step = step_size

    def set_simulation_duration(self, duration: float) -> None:
        self.horizon = duration

    def set_simulation_duration_with_step_size(self, horizon, time_step):
        self.set_simulation_duration(horizon)
        self.time_step = time_step



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""
from typing import Dict, Any, List, Tuple

from apricopt.solving.blackbox.BlackBox import BlackBox
from apricopt.solving.blackbox.BlackBoxSolver import BlackBoxSolver

DEFAULT_MAX_RESTARTS = 1
DEFAULT_MAX_NON_IMPROVING_RESTARTS = 1

custom_options_names = ['method', 'max_restarts', 'max_non_improving_restarts', 'restart_points', 'jac']


class SciPySolver(BlackBoxSolver):

    def solve(self, black_box: BlackBox, solver_params: Dict[str, Any]) \
            -> (Dict[str, float], float, float, float, float):
        # TODO check validity of solver params (e.g. max_restarts > 0)
        from scipy.optimize import minimize

        if 'method' not in solver_params:
            raise ValueError("Please specify an optimization method.")
        max_restarts = DEFAULT_MAX_RESTARTS if 'max_restarts' not in solver_params else solver_params["max_restarts"]

        if max_restarts > 1:
            try:
                restart_points = solver_params["restart_points"]
            except KeyError:
                raise KeyError("If requesting more than one restart, you must provide a list of initial points using "
                               "the option 'restart_points'")
        else:
            restart_points = [black_box.get_np_array_initial_values()]

        max_non_improving_restarts = DEFAULT_MAX_NON_IMPROVING_RESTARTS \
            if 'max_non_improving_restarts' not in solver_params else solver_params['max_non_improving_restarts']

        restart = 0
        non_improving = 0

        best_result = None
        best_result_cost = float('Inf')
        if black_box.get_optimization_parameters_upper_bounds_nparray() is None:
            bounds = None
        else:
            bounds: List[Tuple[float, float]] = []
            param_ids = black_box.get_optimization_parameters_ids()
            for param_ind, param_id in enumerate(param_ids):
                bounds += [(black_box.get_optimization_parameter_lower_bound(param_id),
                            black_box.get_optimization_parameter_upper_bound(param_id))]
        jac_method = None
        if 'jac' in solver_params:
            jac_method = solver_params['jac']
        while restart < max_restarts and non_improving < max_non_improving_restarts:
            print("[SciPySolver:solve] Starting restart {}".format(restart))
            x0 = restart_points[restart]
            result = minimize(black_box.evaluate_objective_np_array, x0, args=(),
                              method=solver_params['method'],
                              bounds=bounds,
                              jac=jac_method,
                              options={k: solver_params[k] for k in solver_params if k not in custom_options_names})
            if result['fun'] < best_result_cost:
                best_result = result
                best_result_cost = result['fun']
                non_improving = 0
                print(
                    "[SciPySolver:solve] Restart {} found a new minimum with cost {}".format(restart, best_result_cost))
            else:
                non_improving += 1
                print("[SciPySolver:solve] Restart {} did not find a new minimum".format(restart))
            restart += 1

        if non_improving == max_non_improving_restarts:
            print("[SciPySolver:solve] Reached maximum number of consecutive restarts without improvements. Stopping.")
        print("[SciPySolver:solve] {} restart{} completed".format(restart,
                                                                  "" if restart == 1 else "s"))
        optimal_params = best_result['x']
        f_return = best_result['fun']
        h_return = 0
        nb_evals = best_result['nfev']
        nb_iters = best_result['nit'] if 'nit' in best_result else 1

        return optimal_params, f_return, h_return, nb_evals, nb_iters



class ObservableFunctionNotFoundError(ValueError):
    def __init__(self, not_found_function_name):
        # Call the base class constructor with the parameters it needs
        super().__init__(f"Observable function with name {not_found_function_name} not found.")
        self.errors = f"Observable function with name {not_found_function_name} not found."



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

import numpy as np

def sphere_array(x):
    return np.square(x).sum(axis=1)

def sphere(x):
    return np.square(x).sum()

def sphere_py(x):
    """
    This method implements the function sum_{i in [1..N]} x^2.
    Its global minimum is 0 in point (0,...,0)
    :param x: a list of float, where the i-th element is the i-th coordinate of the point to test
    :return: a float
    """
    res = 0
    for xi in x:
        res += xi ** 2
    return res



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

from abc import ABC, abstractmethod
from typing import Dict

from apricopt.model.Model import Model
from apricopt.simulation.SimulationEngine import SimulationEngine


class WhiteBoxSolver(ABC):

    def __init__(self):
        self.times = []
        self.objective_values = []
        self.admissible = []
        self.constraints_values = {}
        self.fast_constraints_values = {}
        self.log = []

    @abstractmethod
    def solve(self, model: Model, time_horizon: float, sim_engine: SimulationEngine,
              solver_params: list) -> (Dict[str, float], float, float, float, float):
        pass

    def initialize_storage(self, model: Model) -> None:
        self.times = []
        self.objective_values = []
        self.constraints_values = {}
        self.fast_constraints_values = {}
        for fc in model.constraints:
            self.constraints_values[fc.id] = []
        for fast_c in model.fast_constraints:
            self.fast_constraints_values[fast_c.id] = []
        self.log = []



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

import sympy
import numpy as np
from typing import Dict
import re
from .ObservableFunction import *

replacements = {"^": "**"}


def make_compliant_with_sympy(expressions: List[str]) -> List[str]:
    result: List[str] = list()
    for expr in expressions:
        for old, new in replacements.items():
            expr = expr.replace(old, new)
        expr = expr.strip()
        result.append(expr)
    return result


def get_assignment_at_step(assignment: Dict[str, List[float]], t: int) -> Dict[str, float]:
    assign: Dict[str, float] = dict()
    for v_id, sequence in assignment.items():
        assign[v_id] = sequence[t]
    return assign


def compute_expressions_values(trajectory: Dict[str, List[float]], expressions: List[str], obs_function: Callable) -> \
        List[List[float]]:
    # matches <decimal> e.g. 1, 0.1, 10.23, 1e12, 1.3e-12
    pattern_0 = re.compile(r"((-?\d+(\.\d+)?)(e-?\d+)?)")
    # matches <var_id> <op> <decimal>
    pattern_1 = re.compile(r"([A-Za-z0-9_]+)\s*([+\-*{1,2}/^])\s*((-?\d+(\.\d+)?)(e-?\d+)?)")
    # matches <decimal> <op> <var_id>
    pattern_2 = re.compile(
        r"((-?\d+(\.\d+)?)([e,E]-?\d+)?)\s*([+\-*{1,2}/^])\s*([A-Za-z0-9_]+)")
    # matches <decimal> <op> <var_id>
    pattern_3 = re.compile(r"([A-Za-z0-9_]+)\s*([+\-*{1,2}/^])\s*([A-Za-z0-9_]+)")
    expressions_values: List[List[float]] = []
    traj_length: int = len(trajectory['time'])
    for expr in expressions:
        sympy_expr = sympy.sympify(expr)
        symbols = list(sympy_expr.free_symbols)
        simple = False
        if len(symbols) == 0:
            match_0 = pattern_0.match(expr)
            if match_0.end() == len(expr):
                simple = True
                value = float(match_0.group(1))
                expressions_values.append([value] * traj_length)
        elif len(symbols) == 1 and str(symbols[0]) in trajectory:
            if not any(op in expr for op in ["+", "-", "*", "-", "^", "**"]):
                expressions_values.append(trajectory[str(symbols[0])])
                simple = True
            else:
                match_1 = pattern_1.match(expr)
                match_2 = pattern_2.match(expr)

                variable = ""
                operation = ""
                decimal = float("Nan")
                match_type = 0
                if match_1 and match_1.end() == len(expr):
                    variable = match_1.group(1).strip()
                    operation = match_1.group(2).strip()
                    decimal = float(match_1.group(3).strip())
                    match_type = 1
                elif match_2 and match_2.end() == len(expr):
                    decimal = float(match_2.group(1).strip())
                    operation = match_2.group(5).strip()
                    variable = match_2.group(6).strip()
                    match_type = 2
                if variable != "":
                    simple = True
                    values = np.array(trajectory[str(symbols[0])])
                    if match_type == 1:
                        a = values
                        b = decimal
                    else:
                        a = decimal
                        b = values
                    if operation == "+":
                        expressions_values.append(list(a + b))
                    elif operation == "-":
                        expressions_values.append(list(a - b))
                    elif operation == "*":
                        expressions_values.append(list(a * b))
                    elif operation == "/":
                        expressions_values.append(list(a / b))
                    else:  # if operation in ["^", "**"]:
                        expressions_values.append(list(a ** b))
        elif len(symbols) == 2 and str(symbols[0]) in trajectory and str(symbols[1]) in trajectory:
            match = pattern_3.match(expr)
            if match and match.end() == len(expr):
                simple = True
                var_1 = match.group(1).strip()
                var_2 = match.group(3).strip()
                operation = match.group(2).strip()
                values_1 = np.array(trajectory[var_1])
                values_2 = np.array(trajectory[var_2])
                if operation == "+":
                    expressions_values.append(list(values_1 + values_2))
                elif operation == "-":
                    expressions_values.append(list(values_1 - values_2))
                elif operation == "*":
                    expressions_values.append(list(values_1 * values_2))
                elif operation == "/":
                    expressions_values.append(list(values_1 / values_2))
                elif operation in ["^", "**"]:
                    expressions_values.append(list(values_1 ** values_2))

        if not simple:
            local_assignment: Dict[str, List[float]] = dict()
            for symbol in symbols:
                local_assignment[str(symbol)] = trajectory[str(symbol)]

            if obs_function in [ObservableFunction.Identity, ObservableFunction.LastValue]:
                expressions_values.append([float(
                    sympy.parse_expr(expr, local_dict=get_assignment_at_step(local_assignment, traj_length - 1)))])
            else:
                expressions_values.append(
                    [float(
                        sympy.parse_expr(expr, local_dict=get_assignment_at_step(local_assignment, t))) for t in
                        range(traj_length)])

    return expressions_values


def compute_expressions_values_from_assignment(assignment: Dict[str, float], expressions: List[str],
                                               obs_function: Callable) -> List[List[float]]:
    trajectory: Dict[str, List[float]] = {v_id: [assignment[v_id]] for v_id in assignment}
    trajectory['time'] = [0]
    expressions_values: List[List[float]] = compute_expressions_values(trajectory, expressions, obs_function)
    return expressions_values


class Observable:

    def __init__(self, obs_id: str, name: str, expressions: List[str],
                 function: Callable = ObservableFunction.Identity,
                 lower_bound: float = float("-Inf"), upper_bound: float = float("Inf")):
        self.id: str = obs_id
        self.name: str = name
        self.expressions: List[str] = make_compliant_with_sympy(expressions)
        self.function: Callable = function
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def evaluate(self, trajectory: Dict[str, List[float]]) -> float:
        expressions_values = compute_expressions_values(trajectory, self.expressions, self.function)
        return self.function(*expressions_values)

    def test(self, assignment: Dict[str, float]) -> float:
        test_trajectory: Dict[str, List[float]] = {"time": [0]}
        for param_id, param_value in assignment.items():
            test_trajectory[param_id] = [param_value]
        return self.evaluate(test_trajectory)




"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2022 Marco Esposito, Leonardo Picchiami.
"""

from apricopt.model.Model import Model
from apricopt.simulation.SimulationEngine import SimulationEngine
from apricopt.model.ModelInstance import ModelInstance
from apricopt.simulation.executable.ExecutableModelInstance import ExecutableModelInstance

from abc import abstractmethod
import subprocess as sub


from typing import Dict, List, Type


class ExecutableEngine(SimulationEngine):
    def __init__(self):
        super().__init__()
        self.sim_in_filename = None
        self.sim_out_filename = None
        self.cwd = None
        self.executable_name = None

    def load_model(self, model_filename: str, **exec_info) -> ModelInstance:
        info = list(exec_info.keys())
        self._executable_info_check(info)

        self.executable_name = model_filename
        self.sim_in_filename = exec_info['sim_in_filename']
        self.sim_out_filename = exec_info['sim_out_filename']
        self.cwd = exec_info['cwd']

        model_obj: str = model_filename
        return ExecutableModelInstance(self, model_obj)

    def simulate_trajectory(self, model: Model, horizon: float) -> Dict[str, List[float]]:
        sim_output = self._run_simulation_task(model, horizon, model.observed_outputs)
        return sim_output

    @abstractmethod
    def simulate_trajectory_and_set(self, model: Model, horizon: float, exclude=None) -> Dict[str, List[float]]:
        pass

    @abstractmethod
    def simulate_trajectory_and_get_state(self, model: Model, horizon: float, exclude=None) -> Dict[str, float]:
        pass

    @abstractmethod
    def restore_state(self, model: Model, changed_values: Dict[str, float]) -> None:
        pass

    @abstractmethod
    def read_simulation_output(self) -> Dict[str, List[float]]:
        pass

    @abstractmethod
    def write_simulation_input(self, params_values: Dict[str, float]) -> None:
        pass

    def model_instance_class(self) -> Type[ModelInstance]:
        return ExecutableModelInstance

    @abstractmethod
    def get_simulation_cmd(self, horizon: float, observed_outputs=None) -> str:
        pass

    def _executable_info_check(self, exec_info: List[str]):
        if 'sim_in_filename' not in exec_info:
            raise ValueError("The ExecutableEngine needs the simulation input file to simulate the executable model")

        if 'sim_out_filename' not in exec_info:
            raise ValueError("The ExecutableEngine needs the simulation output file to read "
                             " all simulation output trajectories")

        if 'cwd' not in exec_info:
            raise ValueError("The ExecutableEngine needs the current working directory (CWD) to run the executable")

    def _get_simulation_output(self, model: Model, observed: List[str]) -> Dict[str, List[float]]:
        if not isinstance(model.instance, ExecutableModelInstance):
            raise TypeError("The object in the field 'instance' of a 'Model' object passed to "
                            "'ExecutableEngine::simulate_trajectory' method must have type 'ExecutableModelInstance'.")

        trajectories = self.read_simulation_output()
        sim_output: Dict[str, List[float]] = dict()
        for obs in observed:
            sim_output[obs] = trajectories[obs]
        return sim_output

    def _run_simulation_task(self, model: Model, horizon: float, observed_outputs: List[str]=None) -> Dict[str, List[float]]:
        if not isinstance(model.instance, ExecutableModelInstance):
            raise TypeError("The object in the field 'instance' of a 'Model' object passed to "
                            "'ExecutableEngine::simulate_trajectory' method must have type 'ExecutableModelInstance'.")

        sub.check_output(self.get_simulation_cmd(horizon, observed_outputs), shell=True, cwd=self.cwd)
        return self._get_simulation_output(model, observed_outputs)





"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2022 Sapienza University of Rome, Marco Esposito, Leonardo Picchiami.
"""

from apricopt.model.ModelInstance import ModelInstance

from typing import Dict


class ExecutableModelInstance(ModelInstance):
    def __init__(self, sim_engine, model_obj):
        super().__init__(model_obj)
        self.sim_engine = sim_engine

        self.abs_tol = -1
        self.rel_tol = -1
        self.step_size = -1
        self.duration = -1

    def set_parameters(self, params_values: Dict[str, float]) -> None:
        self.sim_engine.write_simulation_input(params_values)

    def set_parameter(self, param_name: str, value: float) -> None:
        raise NotImplementedError

    def get_parameters_initial_values(self) -> Dict[str, float]:
        return dict()

    def get_compartment_initial_volumes(self) -> Dict[str, float]:
        return dict()

    def set_simulation_configuration(self, abs_tol: float, rel_tol: float, step_size: float) -> None:
        self.abs_tol = abs_tol
        self.rel_tol = rel_tol
        self.step_size = step_size

    def set_simulation_duration(self, duration: float) -> None:
        self.duration = duration

    def set_simulation_duration_with_step_size(self, horizon, time_step):
        self.duration = horizon
        self.step_size = time_step


"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""
from typing import List, Dict

import numpy as np

from apricopt.model.Model import Model
from apricopt.simulation.SimulationEngine import SimulationEngine
from apricopt.solving.blackbox.BlackBox import BlackBox


class SimulableModelBlackBox(BlackBox):

    def __init__(self, sim_engine: SimulationEngine, model: Model, horizon: float):
        if sim_engine.model_instance_class() != model.instance.__class__:
            raise TypeError("The provided SimulationEngine object must be able to simulate the model.")
        self.sim_engine = sim_engine
        self.model = model
        self.horizon = horizon
        self.param_ids = None
        self.eb_constraint_ids = None
        self.pb_constraint_ids = None

    def evaluate(self, parameters: Dict[str, float], check_input=True) -> Dict[str, float]:
        # ME: Force parameters within bounds
        for p_id, p_value in parameters.items():
            if p_value < self.get_optimization_parameter_lower_bound(p_id):
                parameters[p_id] = self.get_optimization_parameter_lower_bound(p_id)
            elif p_value > self.get_optimization_parameter_upper_bound(p_id):
                parameters[p_id] = self.get_optimization_parameter_upper_bound(p_id)
        # Set parameters in the model
        self.model.set_params(parameters)
        # Simulate the model using its simulation engine
        trajectory = self.sim_engine.simulate_trajectory(self.model, self.horizon)

        # Extract the results from the trajectory produced by the simulator
        result: Dict[str, float] = dict()

        result[self.get_objective_id()] = trajectory[self.get_objective_id()][-1]
        for ebc_id in self.get_extreme_barrier_constraints_ids():
            result[ebc_id] = trajectory[ebc_id][-1]
        for pbc_id in self.get_progressive_barrier_constraints_ids():
            result[pbc_id] = trajectory[pbc_id][-1]
        return result

    def evaluate_np_array(self, params: np.array, check_input=False) -> Dict[str, float]:
        param_dict: Dict[str, float] = dict()
        params_ids = self.get_optimization_parameters_ids()
        for i in range(len(params_ids)):
            param_dict[params_ids[i]] = params[i]
        return self.evaluate(param_dict)

    def evaluate_objective_np_array(self, params: np.array, check_input=False) -> float:
        param_dict: Dict[str, float] = dict()
        params_ids = self.get_optimization_parameters_ids()
        for i in range(len(params_ids)):
            param_dict[params_ids[i]] = params[i]
        return self.evaluate_objective(param_dict)

    def is_input_valid(self, parameters: Dict[str, float]) -> bool:
        return True

    def get_optimization_parameters_number(self) -> int:
        return len(self.model.parameters)

    def get_optimization_parameters_ids(self) -> List[str]:
        if not self.param_ids:
            p_ids = [param.id for param in self.model.parameters.values()]
            p_ids.sort()
            self.param_ids = p_ids
        return self.param_ids

    def get_optimization_parameter_lower_bound(self, param_id) -> float:
        return self.model.parameters[param_id].lower_bound

    def get_optimization_parameter_upper_bound(self, param_id) -> float:
        return self.model.parameters[param_id].upper_bound

    def get_optimization_parameters_lower_bounds_nparray(self) -> np.array:
        lb = []
        for param_id in self.get_optimization_parameters_ids():
            lb += [self.model.parameters[param_id].lower_bound]
        return np.array(lb)

    def get_optimization_parameters_upper_bounds_nparray(self) -> np.array:
        ub = []
        for param_id in self.get_optimization_parameters_ids():
            ub += [self.model.parameters[param_id].upper_bound]
        return np.array(ub)

    def get_optimization_parameter_initial_value(self, param_id) -> float:
        return self.model.parameters[param_id].nominal_value

    # TODO implement caching
    def get_optimization_parameters_initial_values(self) -> np.array:
        res: Dict[str, float] = dict()
        for param_id in self.get_optimization_parameters_ids():
            res[param_id] = self.model.parameters[param_id].nominal_value
        return res

    def get_optimization_parameters_initial_values_np_array(self) -> np.array:
        nv = []
        for param_id in self.get_optimization_parameters_ids():
            nv += [self.model.parameters[param_id].nominal_value]
        return np.array(nv)

    def set_optimization_parameters_initial_values(self, param_values: Dict[str, float]) -> None:
        self.model.set_parameters_nominal_values(param_values)

    def get_optimization_parameter_granularity(self, param_id) -> float:
        return self.model.parameters[param_id].granularity

    def get_extreme_barrier_constraints_number(self) -> int:
        return len(self.model.fast_constraints)

    def get_progressive_barrier_constraints_number(self) -> int:
        return len(self.model.constraints)

    def get_extreme_barrier_constraints_ids(self) -> List[str]:
        if not self.eb_constraint_ids:
            c_ids = [constraint.id for constraint in self.model.fast_constraints]
            c_ids.sort()
            self.eb_constraint_ids = c_ids
        return self.eb_constraint_ids

    def get_progressive_barrier_constraints_ids(self) -> List[str]:
        if not self.pb_constraint_ids:
            c_ids = [constraint.id for constraint in self.model.constraints]
            c_ids.sort()
            self.pb_constraint_ids = c_ids
        return self.pb_constraint_ids

    def get_objective_id(self) -> str:
        return self.model.objective.id

    def get_objective_upper_bound(self) -> float:
        return self.model.objective.upper_bound

    @staticmethod
    def get_raisable_exception_type():
        return TypeError

    def finalize(self) -> None:
        pass



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

from typing import List, Dict, Callable, Any

from apricopt.solving.blackbox.BlackBox import BlackBox
from apricopt.solving.blackbox.BlackBoxSolver import BlackBoxSolver
import time


def build_bounds_x0_granularity_lists(black_box: BlackBox) -> (list, list, list):
    bb_params_number: int = black_box.get_optimization_parameters_number()
    lower_bounds: List[float] = [0] * bb_params_number
    upper_bounds: List[float] = [0] * bb_params_number
    x0: List[float] = [0] * bb_params_number

    params_ids: List[str] = black_box.get_optimization_parameters_ids()
    params_ids.sort()  # this can be removed, since it is responsibility of the BB to sort the ids

    granularity_string = "("

    for p_idx in range(bb_params_number):
        lower_bounds[p_idx] = black_box.get_optimization_parameter_lower_bound(params_ids[p_idx])
        upper_bounds[p_idx] = black_box.get_optimization_parameter_upper_bound(params_ids[p_idx])
        x0[p_idx] = black_box.get_optimization_parameter_initial_value(params_ids[p_idx])
        granularity_string += " " + str(black_box.get_optimization_parameter_granularity(params_ids[p_idx]))

    granularity_string += ")"

    return lower_bounds, upper_bounds, x0, granularity_string


def build_output_type_string(black_box: BlackBox) -> str:
    obj = "OBJ"
    eb = " ".join(["EB"] * black_box.get_extreme_barrier_constraints_number())
    pb = " ".join(["PB"] * black_box.get_progressive_barrier_constraints_number())
    return f"{obj} {eb} {pb}"


def build_params_value_dict(black_box: BlackBox, eval_point) -> Dict[str, float]:
    """

    :param black_box:
    :param eval_point: an instance of PyNomad.PyNomadEvalPoint
    :return:
    """
    import PyNomad
    params_values: List[float] = [eval_point.get_coord(i) for i in range(eval_point.size())]
    return build_params_value_dict_from_list(black_box, params_values)


def build_params_value_dict_from_list(black_box: BlackBox, params_values: List[float]) -> Dict[str, float]:
    params_ids: List[str] = black_box.get_optimization_parameters_ids()
    params_ids.sort()  # should be removed

    values_dict: Dict[str, float] = dict()

    for idx in range(len(params_ids)):
        values_dict[params_ids[idx]] = params_values[idx]

    return values_dict


def build_output_string(black_box: BlackBox, sim_output: Dict[str, float]) -> str:
    obj = str(sim_output[black_box.get_objective_id()])

    progressive_constraints_str: str = build_constraints_values_string(
        black_box.get_progressive_barrier_constraints_ids(),
        sim_output)

    extreme_constraints_str: str = build_constraints_values_string(black_box.get_extreme_barrier_constraints_ids(),
                                                                   sim_output)

    return f"{obj} {extreme_constraints_str} {progressive_constraints_str}"


def build_constraints_values_string(constraints_ids: List[str],
                                    sim_output: Dict[str, float]) -> str:
    c: List[str] = list()
    for idx in range(len(constraints_ids)):
        c.append(str(sim_output[constraints_ids[idx]]))
    constraints_string: str = " ".join(c)
    return constraints_string


class NOMADSolver(BlackBoxSolver):

    def __init__(self):
        super().__init__()

    def build_bb_function(self, black_box: BlackBox, print_bb_evals: bool) -> Callable:
        def bb(x, dict_given=False) -> int:

            eval_ok = 1
            sim_output: Dict[str, float]
            try:
                if not dict_given:
                    params_values_dict = build_params_value_dict(black_box, x)
                else:
                    params_values_dict = x

                if print_bb_evals: print(f"\nTrying with params: {params_values_dict}")
                sim_output = black_box.evaluate(params_values_dict)

                line: str

                for extreme_barrier_constraint_id in black_box.get_extreme_barrier_constraints_ids():
                    line = f"\tExtreme barrier constraint '{extreme_barrier_constraint_id}' value: " \
                           f"{sim_output[extreme_barrier_constraint_id]}"
                    if print_bb_evals:
                        print(line, flush=True)

                for progressive_barrier_constraint_id in black_box.get_progressive_barrier_constraints_ids():
                    line = f"\tProgressive barrier constraint '{progressive_barrier_constraint_id}' value: " \
                           f"{sim_output[progressive_barrier_constraint_id]}"
                    if print_bb_evals:
                        print(line, flush=True)

                line = f"\tObjective value: {sim_output[black_box.get_objective_id()]}"
                if print_bb_evals:
                    print(line, flush=True)

                output_string: str = build_output_string(black_box, sim_output)
                x.setBBO(output_string
                         .encode("UTF-8"))

            except Exception as e:
                eval_ok = 0
                sim_output = {}
                print(e)

            if eval_ok == 1:
                self.execution_info(black_box, sim_output)

            return eval_ok

        return bb

    def initialize_storage(self, black_box: BlackBox) -> None:
        super().initialize_storage(black_box)
        self.set_start_time(time.perf_counter())

    def execution_info(self, black_box: BlackBox, sim_output: Dict[str, float]):
        line: str
        extreme_barrier_violated: bool = False
        for extreme_barrier_constraint_id in black_box.get_extreme_barrier_constraints_ids():
            value: float = sim_output[extreme_barrier_constraint_id]
            self.add_extreme_barrier_constraint_value(extreme_barrier_constraint_id, value)
            if value > 0:
                extreme_barrier_violated = True
            line = f"\tExtreme barrier constraint '{extreme_barrier_constraint_id}' value: " \
                   f"{value}"
            self.log.append(line)
        if not extreme_barrier_violated:
            for progressive_barrier_constraint_id in black_box.get_progressive_barrier_constraints_ids():
                value: float = sim_output[progressive_barrier_constraint_id]
                self.add_progressive_barrier_constraint_value(progressive_barrier_constraint_id, value)
                line = f"\tProgressive barrier constraint '{progressive_barrier_constraint_id}' value: " \
                       f"{value}"
                self.log.append(line)

            self.add_objective_value(sim_output[black_box.get_objective_id()])

        self.times.append(time.perf_counter() - self.start_time)

    def solve(self, black_box: BlackBox, solver_params: Dict[str, Any], print_bb_evals=False) -> (Dict[str, float], float, float):
        import PyNomad
        lower_bounds, upper_bounds, x0, granularity_string = build_bounds_x0_granularity_lists(black_box)
        output_type_string = build_output_type_string(black_box)

        black_box_function = self.build_bb_function(black_box, print_bb_evals)

        solve_parameters = [] if "solver_params" not in solver_params else solver_params["solver_params"][:]
        solve_parameters.append(f"BB_OUTPUT_TYPE {output_type_string}")
        solve_parameters.append(" GRANULARITY " + granularity_string)

        print(f"NOMAD parameters: {solve_parameters}", flush=True)

        self.initialize_storage(black_box)

        try:
            #[x_return, f_return, h_return, nb_evals, nb_iters, stopflag] = \
            result = PyNomad.optimize(black_box_function, x0, lower_bounds, upper_bounds, solve_parameters)
            optimal_params = build_params_value_dict_from_list(black_box, result['x_best'])

            black_box.finalize()

            return optimal_params, result['f_best'], result['h_best'], result['nb_evals'], result['nb_iters']
        except Exception as exc:
            print("There was an error invoking the solver:")
            print(exc)
            exit(1)



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

import random

from apricopt.solving.blackbox.FunctionBlackBox import FunctionBlackBox
from apricopt.solving.blackbox.scipy.SciPySolver import SciPySolver
# from apricopt.tests.solving.blackbox.sphere import sphere
from apricopt.tests.solving.blackbox.ackley import ackley
random.seed(14)

dimension = 20

# lower_bounds = [-5.12] * dimension
# upper_bounds = [5.12] * dimension

# Since Nelder-Mead does not support bound constraints, we set them to None
# (we could have called FunctionBlackBox without the last two arguments)
lower_bounds = [-100] * dimension# None
upper_bounds = [100] * dimension# None
#upper_bounds = None
# black_box = FunctionBlackBox(dimension, sphere, lower_bounds, upper_bounds)
black_box = FunctionBlackBox(dimension, ackley, lower_bounds, upper_bounds)

solver = SciPySolver()
max_restarts = 100
max_non_improving_restarts = 10
restart_points = [black_box.get_random_initial_values_np_array() for _ in range(max_restarts)]
options = {"method": "trust-constr",
           'max_restarts': max_restarts,
           'max_non_improving_restarts': max_non_improving_restarts,
           'restart_points': restart_points,
           'jac': '2-point'}

result = solver.solve(black_box, options)

print("\n\nOptimization results:")
print(result)



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

import math
import random
from typing import List, Dict

from apricopt.solving.blackbox.BlackBox import BlackBox


class TestBlackBox(BlackBox):

    def __init__(self, dimensions: int):
        self.d: int = dimensions

    def evaluate(self, parameters: Dict[str, float]) -> Dict[str, float]:
        """
        Implements the Griewank Function, as shown in https://www.sfu.ca/~ssurjano/griewank.html
        :param parameters:
        :return:
        """
        if len(parameters) != self.d:
            raise ValueError

        xx = []
        for i in range(self.d):
            xx.append(parameters[str(i)])
        s = 0
        p = 1
        for ii in range(self.d):
            xi = xx[ii]
            s += xi ** 2 / 4000
            p *= math.cos(xi / math.sqrt(ii + 1))

        result = s - p + 1
        div7 = result % 7.0
        div6 = result % 6.0

        return {"result": result,
                "dummy_extreme_constraint": -div7 if div7 > 0 else 1,
                "dummy_progressive_constraint": -div6 if div6 > 0 else 1}

    def evaluate_and_set_state(self, parameters: Dict[str, float]) -> Dict[str, float]:
        pass

    def set_fixed_parameters(self, params: Dict[str, float]) -> None:
        pass

    def set_optimization_parameters(self, params: Dict[str, float]) -> None:
        pass

    def get_optimization_parameters_number(self) -> int:
        return self.d

    def get_optimization_parameters_ids(self) -> List[str]:
        return [str(i) for i in range(self.d)]

    def get_optimization_parameter_lower_bound(self, param_id) -> float:
        return -600

    def get_optimization_parameter_upper_bound(self, param_id) -> float:
        return 600

    def get_optimization_parameter_initial_value(self, param_id) -> float:
        return random.randint(-600, 600)

    def get_optimization_parameter_granularity(self, param_id) -> float:
        return 0

    def get_extreme_barrier_constraints_number(self) -> int:
        return 1

    def get_progressive_barrier_constraints_number(self) -> int:
        return 1

    def get_extreme_barrier_constraints_ids(self) -> List[str]:
        return ["dummy_extreme_constraint"]

    def get_progressive_barrier_constraints_ids(self) -> List[str]:
        return ["dummy_progressive_constraint"]

    def get_objective_id(self) -> str:
        return "result"

    def get_objective_upper_bound(self):
        return float('Inf')

    @staticmethod
    def get_raisable_exception_type():
        return ValueError

    def set_optimization_parameters_initial_values(self, param_values: Dict[str, float]) -> None:
        pass

    def finalize(self) -> None:
        pass



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""


import COPASI
from apricopt.solving.whitebox.COPASI.COPASIOptimisationMethod import COPASIOptimisationMethod
from apricopt.solving.whitebox.COPASI.COPASISolverParameter import COPASISolverParameter
from typing import List


class COPASITruncatedNewton(COPASIOptimisationMethod):
    def __init__(self):
        super(COPASITruncatedNewton, self).__init__()

    def set_parameters_configuration(self, opt_task: COPASI.COptTask,
                                     solver_parameters: List[COPASISolverParameter] = None) -> None:
        solver_params = self.solver_parameters if solver_parameters is None else solver_parameters
        self.set_optimisation_task(opt_task)
        self.opt_task.setMethodType(COPASI.CTaskEnum.Method_TruncatedNewton)
        self.set_and_check_solver_parameters(solver_params)


"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""
from typing import Dict, Any

from apricopt.solving.blackbox.BlackBox import BlackBox
from apricopt.solving.blackbox.BlackBoxSolver import BlackBoxSolver
import numpy as np

DEFAULT_N_PARTICLES = 10
DEFAULT_ITERATIONS = 1000
DEFAULT_SEARCH_N_PARTICLES = 10
DEFAULT_SEARCH_ITERATIONS = 10
DEFAULT_SEARCH_N_SELECTION_ITERS = 100

def random_search(black_box: BlackBox, n_particles=10, iters=10, n_selection_iters=100):
    print("Starting random search for hyperparameter optimization")
    from pyswarms.utils.search.random_search import RandomSearch
    from pyswarms.single.global_best import GlobalBestPSO
    options = {'c1': [1, 5],
               'c2': [1, 10],
               'w': [2, 5],
               'p': [1, 2],
               'k': [3, 10]}
    g = RandomSearch(GlobalBestPSO, n_particles=n_particles, dimensions=black_box.get_optimization_parameters_number(),
                     options=options, objective_func=black_box.evaluate_objective_multiple_inputs_np_array,
                     iters=iters, n_selection_iters=n_selection_iters, bounds=(black_box.get_optimization_parameters_lower_bounds_nparray(),
                  black_box.get_optimization_parameters_upper_bounds_nparray()))
    best_score, best_options = g.search()
    return best_options


class PySwarmsSolver(BlackBoxSolver):
    def solve(self, black_box: BlackBox, solver_params: Dict[str, Any]) -> (
            Dict[str, float], float, float, float, float):
        import pyswarms as ps
        n_particles = DEFAULT_N_PARTICLES if "n_particles" not in solver_params else solver_params["n_particles"]
        iterations = DEFAULT_ITERATIONS if "iterations" not in solver_params else solver_params["iterations"]

        bounds = (black_box.get_optimization_parameters_lower_bounds_nparray(),
                  black_box.get_optimization_parameters_upper_bounds_nparray())

        if "init_pos" in solver_params.keys():
            inits = solver_params['init_pos']
            init_list = []
            for start in inits:
                init_list.append(
                    [start[param] for param in black_box.get_optimization_parameters_ids()]
                )
            init_pos = np.asarray(init_list, dtype=np.double)

        else:
            init_pos = None

        # Perform hyperparameter optimization if requested, via random search
        if "search" in solver_params and solver_params["search"]:
            search_n_particles = DEFAULT_SEARCH_N_PARTICLES if "search_n_particles" not in solver_params \
                else solver_params["search_n_particles"]
            search_iters = DEFAULT_SEARCH_ITERATIONS if "search_iters" not in solver_params \
                else solver_params["search_iters"]
            search_n_selection_iters = DEFAULT_SEARCH_N_SELECTION_ITERS if \
                "search_n_selection_iters" not in solver_params else solver_params["search_n_selection_iters"]

            options = random_search(black_box, n_particles=search_n_particles, iters=search_iters,
                                    n_selection_iters=search_n_selection_iters)
        else:
            options = solver_params



        optimizer = ps.single.GlobalBestPSO(
            n_particles=n_particles,
            dimensions=black_box.get_optimization_parameters_number(),
            options=options,
            bounds=bounds,
            init_pos=init_pos)

        # Avoids checking whether the variables are inside their bounds at each evaluation of the black-box
        kwargs = {"check_input": False}

        # Perform optimization
        cost, best_x = optimizer.optimize(black_box.evaluate_objective_multiple_inputs_np_array,
                                          iters=iterations, **kwargs)

        optimal_params = best_x
        f_return = cost
        h_return = 0
        nb_evals = iterations * n_particles
        nb_iters = iterations

        return optimal_params, f_return, h_return, nb_evals, nb_iters



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

import random

from apricopt.solving.blackbox.FunctionBlackBox import FunctionBlackBox
from apricopt.solving.blackbox.pyswarms.PySwarmsSolver import PySwarmsSolver
from apricopt.tests.solving.blackbox.ackley import ackley

random.seed(14)

dimension = 20

lower_bounds = [-32.768] * dimension
upper_bounds = [32.768] * dimension

black_box = FunctionBlackBox(dimension, ackley, lower_bounds, upper_bounds)

solver = PySwarmsSolver()

options = {'iterations': 1000, 'n_particles': 10,
           'search': True,
           'search_n_particles': 10,
           'search_iters': 10,
           'search_n_selection_iters': 100}
result = solver.solve(black_box, options)

print(result)



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

from abc import abstractmethod, ABC
from typing import Dict, List

import numpy as np


class BlackBox(ABC):
    @abstractmethod
    def evaluate(self, parameters: Dict[str, float], check_input=True) -> Dict[str, float]:
        """
        Evaluate the black box over the given assignment of the parameters.
        :param check_input:
        :param parameters:
        :return:
        """

    def evaluate_objective(self, parameters: Dict[str, float], check_input=False) -> float:
        """

        :param check_input:
        :param parameters:
        :return:
        """
        return self.evaluate(parameters, check_input)[self.get_objective_id()]

    def evaluate_multiple_inputs(self, parameters_list: List[Dict[str, float]], check_input=False) \
            -> List[Dict[str, float]]:
        return [self.evaluate(params, check_input) for params in parameters_list]

    def evaluate_objective_multiple_inputs(self, parameters_list: List[Dict[str, float]], check_input=False) \
            -> List[float]:
        return [self.evaluate_objective(params, check_input) for params in parameters_list]

    def evaluate_multiple_inputs_np_array(self, parameters_list: np.ndarray, check_input=False) \
            -> List[Dict[str, float]]:
        results: List[Dict[str, float]] = [{}] * len(parameters_list)
        for i in range(len(parameters_list)):
            results[i] = self.evaluate_np_array(parameters_list[i], check_input)
        return results
        # return [self.evaluate_np_array(params, check_input) for params in parameters_list]

    def evaluate_objective_multiple_inputs_np_array(self, parameters_list: np.ndarray, check_input=False) \
            -> List[float]:
        results: List[float] = [0] * len(parameters_list)
        for i in range(len(parameters_list)):
            results[i] = self.evaluate_objective_np_array(parameters_list[i], check_input)
        return results
        # return [self.evaluate_objective_np_array(params, check_input) for params in parameters_list]

    @abstractmethod
    def evaluate_np_array(self, parameters: np.array, check_input=False) -> Dict[str, float]:
        """

        :param check_input:
        :param parameters:
        :return:
        """
        pass

    @abstractmethod
    def evaluate_objective_np_array(self, parameters: np.array, check_input=False) -> float:
        """

        :param check_input:
        :param parameters:
        :return:
        """
        pass

    @abstractmethod
    def is_input_valid(self, parameters: Dict[str, float]) -> bool:
        """
        Checks whether the given parameters are a valid input for the black-box
        :param parameters: the parameters to check
        :return: true if the parameters are valid, false otherwise
        """
        pass

    @abstractmethod
    def get_optimization_parameters_number(self) -> int:
        """
        Returns the number of parameters to be optimized.
        :return: an integer representing the number of optimization parameters.
        """
        pass

    @abstractmethod
    def get_optimization_parameters_ids(self) -> List[str]:
        """
        Returns the identifiers of the optimization parameters as a list of strings.
        The order in which the identifiers appear in the list should be consistent among multiple calls.
        :return: a list of strings representing the identifiers of the optimization parameters, in some fixed order.
        """
        pass

    @abstractmethod
    def get_optimization_parameter_lower_bound(self, param_id) -> float:
        """
        Returns the lower bound of the optimization parameter with the given identifier.
        :param param_id: the identifier of the parameter.
        :return: a float representing the lower bound of the optimization parameter with the given identifier.
        """
        pass

    @abstractmethod
    def get_optimization_parameter_upper_bound(self, param_id) -> float:
        """
        Returns the upper bound of the optimization parameter with the given identifier.
        :param param_id: the identifier of the parameter.
        :return: a float representing the upper bound of the optimization parameter with the given identifier.
        """
        pass

    @abstractmethod
    def get_optimization_parameters_lower_bounds_nparray(self) -> np.array:
        """
        Returns the lower bounds of the optimization parameters as a numpy array.
        :return: a numpy array containing the lower bounds of the optimization parameters in the order of their ids.
        """
        pass

    @abstractmethod
    def get_optimization_parameters_upper_bounds_nparray(self) -> np.array:
        """
        Returns the upper bounds of the optimization parameters as a numpy array.
        :return: a numpy array containing the upper bounds of the optimization parameters in the order of their ids.
        """
        pass

    @abstractmethod
    def get_optimization_parameter_initial_value(self, param_id) -> float:
        """
        Returns the initial_value of the optimization parameter with the given identifier.
        :param param_id: the identifier of the parameter.
        :return: a float representing the initial_value of the optimization parameter with the given identifier.
        """
        pass

    def get_optimization_parameters_initial_values(self) -> Dict[str, float]:
        return {p_id: self.get_optimization_parameter_initial_value(p_id)
                for p_id in self.get_optimization_parameters_ids()}

    @abstractmethod
    def set_optimization_parameters_initial_values(self, param_values: Dict[str, float]) -> None:
        """
        Sets the initial_values of the optimization parameters in the dictionary with the given values.
        :param param_values: a dictionary that maps valid parameter identifiers to values.
        :return: None.
        """
        pass

    def get_np_array_initial_values(self) -> np.array:
        x0 = self.get_optimization_parameters_initial_values()
        x0_list = []
        for param_id in self.get_optimization_parameters_ids():
            x0_list.append(x0[param_id])
        return np.array(x0_list)

    @abstractmethod
    def get_optimization_parameter_granularity(self, param_id) -> float:
        """
        Returns the granularity of the optimization parameter with the given identifier.
        If the parameter has no granularity, i.e. it should be optimized over a continuous domain, this
        method should return 0 (zero).
        :param param_id: the identifier of the parameter.
        :return: a float representing the granularity of the optimization parameter with the given identifier. If
        the parameter has no granularity, returns 0.
        """
        pass

    @abstractmethod
    def get_extreme_barrier_constraints_number(self) -> int:
        """
        Returns the number of extreme barrier constraints, i.e. the number of constraints that define the optimization
        space. If any of the extreme barrier constraints is violated, the point is considered as invalid by the solver
        in the optimization.
        :return: a non-negative int representing the number of extreme barrier constraints.
        """
        pass

    @abstractmethod
    def get_progressive_barrier_constraints_number(self) -> int:
        """
        Returns the number of progressive barrier constraints, i.e. the number of constraints that define the space
        of feasible solutions. If a progressive barrier constraints is violated, the point is still considered as valid
        by the solver and used in the optimization process.
        :return: a non-negative int representing the number of progressive barrier constraints.
        """
        pass

    @abstractmethod
    def get_extreme_barrier_constraints_ids(self) -> List[str]:
        """
        Returns the identifiers of extreme barrier constraints, i.e. the constraints that define the optimization
        space. If any of the extreme barrier constraints is violated, the point is considered as invalid by the solver
        in the optimization.
        The order in which the identifiers are returned in the list must be consistent among all method calls.
        :return: a list of strings representing the identifiers of extreme barrier constraints.
        """
        pass

    @abstractmethod
    def get_progressive_barrier_constraints_ids(self) -> List[str]:
        """
        Returns the identifiers of progressive barrier constraints, i.e. the constraints that define the space
        of feasible solutions. If a progressive barrier constraints is violated, the point is still considered as valid
        by the solver and used in the optimization process.
        The order in which the identifiers are returned in the list must be consistent among all method calls.
        :return: a list of strings representing the identifiers of progressive barrier constraints.
        """
        pass

    @abstractmethod
    def get_objective_id(self) -> str:
        """
        Returns an identifier for the objective value. It must be different from any of the identifiers
        of the parameters.
        :return: A string representing an identifier for the objective value.
        """
        pass

    @abstractmethod
    def get_objective_upper_bound(self) -> float:
        pass

    @staticmethod
    @abstractmethod
    def get_raisable_exception_type():
        return Exception

    @abstractmethod
    def finalize(self) -> None:
        """
        Performs all needed operations to finalize the black box.
        This method is invoked by BlackBoxSolver at the end of the optimization.
        :return:
        """
        pass

    def force_parameter_value_in_bounds(self, param_id, value):
        return min(
            self.get_optimization_parameter_upper_bound(param_id),
            max(
                self.get_optimization_parameter_lower_bound(param_id),
                value))



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2022 Sapienza University of Rome, Marco Esposito, Leonardo Picchiami.
"""
import math

from apricopt.model.Model import Model
from apricopt.model.Parameter import Parameter
from apricopt.simulation.SimulationEngine import SimulationEngine
from typing import Dict, List, Collection
import numpy as np
from scipy import stats


def sample_parameters(parameters: Collection[Parameter]) -> Dict[str, float]:
    result: Dict[str, float] = dict()
    samples = np.random.rand(len(parameters))
    for i, param in enumerate(parameters):
        if not param.optimize:
            # parameter is fixed (e.g. measured)
            result[param.id] = param.nominal_value
            continue
        if param.distribution == 'uniform':
            result[param.id] = \
                round(param.lower_bound + (
                        param.upper_bound - param.lower_bound) * samples[i])
        elif param.distribution == 'loguniform':
            result[param.id] = 10 ** \
                                         (math.log(param.lower_bound, 10) + (
                                                 math.log(param.upper_bound, 10) - math.log(param.lower_bound,
                                                                                            10))
                                          * samples[i])
        elif param.distribution == 'lognormal':
            result[param.id] = stats.lognorm(param.sigma, scale=param.mu).ppf(
                samples[i])
    return result


class Sampler:
    def __init__(self, model: Model, sim_engine: SimulationEngine, horizon: float, seed: int):
        self.model = model
        self.sim_engine = sim_engine
        self.horizon = horizon
        self.seed = seed
        np.random.seed(seed)

    def random_sampling_admissible_parameter_with_respect_constraints(self) -> Dict[str, float]:
        sampled_parameters = self.parameters_random_sampling()
        self.model.set_fixed_params(sampled_parameters)
        evaluated_constraints = self.sim_engine.simulate(self.model, self.horizon)
        admissible = all(evaluated_constraints[constraint.id] <= 0 for constraint in self.model.constraints)
        while not admissible:
            sampled_parameters = self.parameters_random_sampling()
            self.model.set_fixed_params(sampled_parameters)
            evaluated_constraints = self.sim_engine.simulate(self.model, self.horizon)
            admissible = all(evaluated_constraints[constraint.id] <= 0 for constraint in self.model.constraints)
        return sampled_parameters

    def random_sampling_admissible_parameter_with_respect_fast_constraints(self) -> Dict[str, float]:
        pass

    def random_sampling_admissible_parameter_with_respect_all_constraints(self) -> Dict[str, float]:
        pass

    def parameters_random_sampling(self) -> Dict[str, float]:
        parameters: List[Parameter] = list(self.model.parameters.values())
        parameters.sort(key=lambda x: x.id)
        sample = np.random.rand(len(parameters))
        sampled_parameters: Dict[str, float] = dict()
        for i in range(len(parameters)):
            param = parameters[i]
            if param.distribution == 'uniform':
                sampled_parameters[param.id] = \
                    round(param.lower_bound + (
                            param.upper_bound - param.lower_bound) * sample[i])
        return sampled_parameters




"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

from apricopt.model.Model import Model
from apricopt.simulation.COPASI.COPASIEngine import COPASIEngine
from apricopt.model.Observable import Observable
from apricopt.model.Parameter import Parameter
from apricopt.solving.whitebox.COPASI.COPASISolver import COPASISolver
from apricopt.solving.whitebox.COPASI.method.COPASIGeneticAlgorithm import COPASIGeneticAlgorithm
from apricopt.solving.whitebox.COPASI.COPASISolverParameter import COPASISolverParameter


def define_optimisation_problem(model: Model) -> None:
    '''
    In this function the optimisation problem modeled in SBML is connected with APRICOPT.

    Objective function: such function can be modeled in SBML either as specie or parameter and it must be connected
     through an Observable. It is needed that the obs_id is the respective SBML ID of modeled objective function.

    Optimisation parameters: they can be modeled in SBML either as species or parameters and they must be connected
     through a Parameter. It is needed that the param_id is the respective SBML ID of modeled parameter.
     The inputs upper_bound and lower_bound are needed to define the upper e lower bounds for each optimisation variable.
     If you do not set such values, APRICOPT will set the bounds as, respectively, +Inf and -Inf. The input nominal_value
     is used as x0 value for the specific variable.

    Constraints: they can be modeled in SBML either as species or parameters and they must be connected through an
     Observable. It is needed that the obs_id is the respective SBML ID of modeled objective function.
     The upper_bound and lower_bound are needed to specify the contexts in which a certain constraint is violated.
     If you do not set such values, APRICOPT will set the bounds as, respectively, +Inf and -Inf. In COPASI,
     the constaints are managed by an upper bound an a lower bound and such constraint can assume only the values
     within the bounds (bounds included). If you do not set the nominal value, APRICOPT will set such value as +Inf.

    :param model:
    :return:
    '''

    model.objective = Observable(obs_id='objective', name='objective_function', expressions=["0"])

    model.parameters['x1'] = Parameter(param_id='x1', name='x1 value', lower_bound=-1.5, upper_bound=1.5, nominal_value=0.5)
    model.parameters['x2'] = Parameter(param_id='x2', name='x2 value', lower_bound=-0.5, upper_bound=2.5,
                                       nominal_value=1.0)

    model.constraints.add(
        Observable(obs_id='x1_x2_constraint', name='x1 x2 constraint',
                   expressions=["0"], lower_bound=-1, upper_bound=0)
    )

    model.constraints.add(
        Observable(obs_id='x1_x2_constraint_2', name='x1 x2 constraint',
                   expressions=["0"], lower_bound=-1, upper_bound=0)
    )


def main() -> None:
    '''
    This file shows how to minimise (just to have an example) the Rosenbrock function with a SBML-based problem modeling.
    Note that, by default, the solver performs a minimisation.

    The optimisation through COPASI Solver is performed as follow:
      1. Create a COPASI Simulation Engine a Model that contains your model with the related simulation configurations.
         In general, different solvers can use different simulation engine. In such context, the COPASI solver can use
         only the COPASI simulation engine.
      2. Create the instance of the chosen optimisation method and then create the COPASI Solver with the chosen
         method and the optimisation log pathname.
      3. Create a list of solver parameters through a COPASISolverParameter object. Note that different solvers can
         have different solver parameters.
      4. Optimise through the solve method.

    :return:
    '''

    sim_engine = COPASIEngine()
    model = Model(sim_engine, "rosenbrock_sbml.xml", abs_tol=1e-12, rel_tol=1e-6, time_step=1)
    define_optimisation_problem(model)

    method = COPASIGeneticAlgorithm()
    solver = COPASISolver(method, "rosenbrock_optimisation_min.log")
    solver_parameters = [
        COPASISolverParameter(COPASIGeneticAlgorithm.NUMBER_OF_GENERATIONS, 40),
        COPASISolverParameter(COPASIGeneticAlgorithm.POPULATION_SIZE, 40),
        COPASISolverParameter(COPASIGeneticAlgorithm.MUTATION_VARIANCE, 0.1),
        COPASISolverParameter(COPASIGeneticAlgorithm.SEED, 1)
    ]


    time_horizon = 100
    optimized_parameters, objective_value, \
        function_evaluations, statistics = solver.solve(model, time_horizon, sim_engine,
                                                        solver_params=solver_parameters)

    x1 = 0.5
    x2 = 1
    a = 1
    b = 100
    rosen_x0 = (a - x1)**2 + b*(x2 - x1**2)**2
    print("==== APRICOPT - COPASI Solver example usage: Rosenbrock ====\n")
    print(f"The objective function evaluated at x0 point: {rosen_x0:.2f}")
    print(f"the objective function value after {function_evaluations} function evaluations: {objective_value:.2f}")
    print(f"The optimised parameters are:\n\t\t x1: {optimized_parameters['x1']}\n\t\t x2: {optimized_parameters['x2']}")


if __name__ == '__main__':
    main()


"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

from typing import Dict, Set, Tuple, Callable, List

import re, yaml
from pandas import DataFrame

from apricopt.model.FastObservable import FastObservable
from apricopt.model.LogObservable import LogObservable
from apricopt.model.Observable import Observable
from apricopt.model.ObservableFunction import ObservableFunction
from apricopt.model.Parameter import Parameter


def get_parameter_space(parameter_df: DataFrame) -> Set[Parameter]:
    params: Set[Parameter] = set()
    for param_id, data in parameter_df.iterrows():
        params.add(Parameter(str(param_id), data.parameterName,
                             data.lowerBound, data.upperBound, data.nominalValue,
                             data.distribution, data.mu, data.sigma, data.granularity, data.estimate))
    return params


def get_observable_formula(formula: str) -> Tuple[Callable, List[str]]:
    pattern = re.compile(r"(^[A-Za-z0-9_]+)\(([^(]+)\)$")
    match = pattern.match(formula)
    if not match:
        return ObservableFunction.Identity, [formula]
    function_name, expressions_str = match.groups()
    obs_function = ObservableFunction.get_function(function_name)
    if not obs_function:
        return ObservableFunction.Identity, [formula]
    expressions: List[str] = [s.strip() for s in expressions_str.split(',')]
    return obs_function, expressions


def get_objective(objective_df: DataFrame) -> Observable:
    data = objective_df.iloc[0]

    func, expressions = get_observable_formula(data.observableFormula)

    result = Observable(data.name, data.observableName, expressions, function=func,
                        lower_bound=data.lowerBound, upper_bound=data.upperBound)
    return result


def get_constraints(constraints_df: DataFrame) -> Set[Observable]:
    result: Set[Observable] = set()
    for obj_id, data in constraints_df.iterrows():
        func, expressions = get_observable_formula(data.observableFormula)
        result.add(Observable(str(obj_id), data.observableName, expressions, function=func))

    return result


def get_log_observables(log_observables_df: DataFrame) -> Set[LogObservable]:
    result: Set[LogObservable] = set()
    for obj_id, data in log_observables_df.iterrows():
        func, expressions = get_observable_formula(data.observableFormula)
        result.add(LogObservable(str(obj_id), data.observableName, expressions, func, data.message, data.formatMode))
    return result


def get_fast_constraints(constraints_df: DataFrame) -> Set[FastObservable]:
    result: Set[FastObservable] = set()
    for obj_id, data in constraints_df.iterrows():
        func, expressions = get_observable_formula(data.observableFormula)
        result.add(FastObservable(str(obj_id), data.observableName, expressions, function=func))
    return result


def get_conditions(conditions_df: DataFrame) -> Dict[str, Dict[str, float]]:
    cds: Dict[str, Dict[str, float]] = dict()

    for cd_id, params in conditions_df.iterrows():
        cd_id_str = str(cd_id)
        cds[cd_id_str] = dict()
        for name, value in params.items():
            cds[cd_id_str][name] = value
    return cds


def parse_config_file(config_filename: str) -> dict:
    f = open(config_filename)
    data: dict = yaml.load(f, Loader=yaml.FullLoader)
    f.close()
    return data


# Needed for COPASI Solver: it could be merged into the get_constraints function
def get_constraints_with_bounds(constraints_df: DataFrame) -> Set[Observable]:
    result: Set[Observable] = set()
    for obj_id, data in constraints_df.iterrows():
        func, expressions = get_observable_formula(data.observableFormula)
        result.add(Observable(str(obj_id), data.observableName, expressions, function=func, lower_bound=data.lowerBound,
                              upper_bound=data.upperBound))

    return result



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020 Marco Esposito, Leonardo Picchiami.
"""

from typing import Dict, List, Type

from apricopt.model.Model import Model
from apricopt.model.ModelInstance import ModelInstance
from apricopt.simulation.MockUp.MockUpModelInstance import MockUpModelInstance
from apricopt.simulation.SimulationEngine import SimulationEngine

import random


class MockUpSimulationEngine(SimulationEngine):

    def __init__(self):
        super().__init__()

    def load_model(self, model_filename: str) -> ModelInstance:
        return MockUpModelInstance(model_filename)

    def simulate_trajectory(self, model: Model, horizon: float) -> Dict[str, List[float]]:
        return {"time": list(range(int(horizon))),
                "output": [random.random() for _ in range(int(horizon))]}

    def simulate_trajectory_and_set(self, model: Model, horizon: float, exclude=None) -> Dict[str, List[float]]:
        return self.simulate_trajectory(model, horizon)

    def model_instance_class(self) -> Type[ModelInstance]:
        return MockUpModelInstance

    def simulate_trajectory_and_get_state(self, model: Model, horizon: float, exclude=None) -> Dict[str, List[float]]:
        return self.simulate_trajectory(model, horizon)

    def restore_state(self, model: Model, changed_values: dict) -> None:
        pass



from typing import List, Callable, Dict

from apricopt.model.Observable import Observable, compute_expressions_values_from_assignment, compute_expressions_values


class LogObservable(Observable):

    def __init__(self, param_id: str, name: str, expression: List[str], function: Callable,
                 message: str, format_mode: str = "append"):
        super(LogObservable, self).__init__(param_id, name, expression, function)
        self.message: str = message
        self.format_mode: str = format_mode

    def evaluate(self, trajectory: Dict[str, List[float]]) -> str:
        expressions_values = compute_expressions_values(trajectory, self.expressions, self.function)
        value: float = self.function(*expressions_values)
        return self.format_message(value)

    def value(self, trajectory: Dict[str, List[float]]) -> float:
        expressions_values = compute_expressions_values(trajectory, self.expressions, self.function)
        return self.function(*expressions_values)

    def format_message(self, value: float) -> str:
        if self.format_mode == "append":
            return self.message + str(value)
        else:
            return self.message.format(value)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="apricopt",
    version="0.0.2a3dev12",
    author="Marco Esposito",
    author_email="esposito@di.uniroma1.it",
    description="A library for simulation-based parameter optimization",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://mclab.di.uniroma1.it",
    packages=setuptools.find_packages(),
    install_requires=[
        'attrs==21.2.0',
        'chaospy==4.2.1',
        'cycler==0.10.0',
        'future==0.18.2',
        'importlib-metadata==3.1.0',
        'kiwisolver==1.3.2',
        'matplotlib==3.5.0',
        'mpmath==1.2.1',
        'numpy==1.22.3',
        'Pillow==8.3.2',
        'pyparsing==2.4.7',
        'python-dateutil==2.8.2',
        'python-libsbml==5.19.5',
        'PyYAML==5.4.1',
        'scipy==1.8.1',
        'seaborn==0.11.0',
        'six==1.16.0',
        'sympy==1.8',
        'tqdm==4.62.2',
        'zipp==3.4.0',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    
    python_requires='>=3.6',
)



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

import math

import numpy as np


def ackley(x) -> float:
    """
    The Ackley function
    :param x: the input
    :return: a float
    """
    a = 20
    b = 0.2
    c = 2 * math.pi

    d = len(x)
    sum1 = np.square(x).sum()#axis=1)
    dot = np.dot(c, x)
    cos = np.cos(dot)
    sum2 = cos.sum()#axis=1)

    term1 = -a * np.power(math.e, -b * np.sqrt(sum1 / d))
    term2 = -np.power(math.e, sum2 / d)

    return term1 + term2 + a + math.e

def ackley_py(x) -> float:
    """
    The Ackley function
    :param x: the input
    :return: a float
    """
    d = len(x)
    sum1 = 0
    sum2 = 0
    for xi in x:
        sum1 += xi**2
        sum2 += math.cos(2 * math.pi*xi)

    return -20 * math.pow(math.e, -0.2*math.sqrt(sum1 / d)) + -math.pow(math.e, sum2 / d) + 20 + math.e



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2022 Sapienza University of Rome, Marco Esposito, Leonardo Picchiami.
"""

from apricopt.model.Model import Model
from apricopt.simulation.executable.ExecutableEngine import ExecutableEngine

from typing import Dict, List


class ModelicaExecutableEngine(ExecutableEngine):

    def simulate_trajectory_and_set(self, model: Model, horizon: float, exclude=None) -> Dict[str, List[float]]:
        raise NotImplementedError

    def simulate_trajectory_and_get_state(self, model: Model, horizon: float, exclude=None) -> Dict[str, float]:
        raise NotImplementedError

    def restore_state(self, model: Model, changed_values: Dict[str, float]) -> None:
        raise NotImplementedError

    def read_simulation_output(self) -> Dict[str, List[float]]:
        results: Dict[str, List[float]] = dict()
        with open(f'{self.cwd}/{self.sim_out_filename}', 'r') as f:
            for line in f.readlines():
                words = line.split()
                results[words[0]] = [float(words[1])]
        return results

    def write_simulation_input(self, params_values: Dict[str, float]) -> None:
        start = True
        with open(f'{self.cwd}/{self.sim_in_filename}', 'w') as f:
            for name, val in params_values.items():
                if start:
                    f.write(f'{name}={val}')
                    start = False
                else:
                    f.write('\n')
                    f.write(f'{name}={val}')

    def get_simulation_cmd(self, horizon: float, observed_outputs=None) -> str:
        return f'./{self.executable_name} -overrideFile={self.sim_in_filename}'


"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

import COPASI
from abc import ABC, abstractmethod
from typing import List, Union
from apricopt.solving.whitebox.COPASI.COPASISolverParameter import COPASISolverParameter


class COPASIOptimisationMethod(ABC):
    def __init__(self):
        self.opt_task: Union[COPASI.COptTask, None] = None
        self.solver_parameters: List[COPASISolverParameter] = []

    def set_and_check_solver_parameters(self, solver_parameters: List[COPASISolverParameter]):
        if self.opt_task is None:
            raise ValueError("Optimisation Task has not been set")

        opt_method = self.opt_task.getMethod()
        if opt_method is None:
            raise ValueError("A valid optimisation method has not been set")

        for param in solver_parameters:
            copasi_param = opt_method.getParameter(param.id)
            if copasi_param is None:
                raise ValueError(f"The parameter {param.id} is wrong for such optimisation method")

            if copasi_param.getType() == COPASI.CCopasiParameter.Type_UINT or \
                                            copasi_param.getType() == COPASI.CCopasiParameter.Type_INT:
                copasi_param.setIntValue(int(param.value))
            elif copasi_param.getType() == COPASI.CCopasiParameter.Type_DOUBLE:
                copasi_param.setDblValue(float(param.value))
            elif copasi_param.getType() == COPASI.CCopasiParameter.Type_UDOUBLE:
                copasi_param.setUDblValue(float(param.value))

    def set_optimisation_task(self, opt_task: COPASI.COptTask):
        self.opt_task = opt_task

    @abstractmethod
    def set_parameters_configuration(self, opt_task: COPASI.COptTask,
                                     solver_parameters: List[COPASISolverParameter] = None) -> None:
        pass



from typing import List

from apricopt.IO.data_input import get_observable_formula
from apricopt.model.Model import Model
from apricopt.model.Observable import Observable
from apricopt.model.ObservableFunction import ObservableFunction
import math

from apricopt.model.Parameter import Parameter


def test_user_function(*args: List[float]) -> float:
    return math.cos(args[0][0]) + math.sin(args[1][0])


ObservableFunction.add_function("TestFunction", test_user_function)
print(ObservableFunction.get_function_names())

m = Model(sim_engine=None, model_filename="", abs_tol=0, rel_tol=0, time_step=0)
func, expressions = get_observable_formula("TestFunction(p1,p2)")

m.objective = Observable("obj", "objName", expressions, function=func,
                         lower_bound=-2, upper_bound=2)

p1 = Parameter("p1", "param1", -1, 1, 0)
p2 = Parameter("p2", "param2", -1, 1, 0)

m.set_parameter_space({p1, p2})

for v1 in [0, math.pi/6, math.pi/4, math.pi/3, math.pi/2, math.pi/2+math.pi/6, math.pi/2+math.pi/3, math.pi]:
    for v2 in [0, math.pi/6, math.pi/4, math.pi/3, math.pi/2, math.pi/2+math.pi/6, math.pi/2+math.pi/3, math.pi]:
        print(f"obj({v1}, {v2}) = {m.objective.test({'p1': v1, 'p2': v2})}")



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2022 Sapienza University of Rome, Marco Esposito, Leonardo Picchiami.
"""

from abc import ABC, abstractmethod
from typing import Dict, List, Type, Set

from apricopt.model.Model import Model
from apricopt.model.ModelInstance import ModelInstance
from apricopt.model.Observable import Observable


class SimulationEngine(ABC):

    def __init__(self):
        pass

    def simulate(self, model: Model, horizon: float) -> Dict[str, float]:
        trajectory: Dict[str, List[float]] = self.simulate_trajectory(model, horizon)
        result = model.evaluate_constraints(trajectory)
        return result

    def simulate_and_set(self, model: Model, horizon: float, exclude=None, evaluate_constraints=True) -> Dict[str, float]:
        if exclude is None:
            exclude = []
        trajectory = self.simulate_trajectory_and_set(model, horizon, exclude=exclude)

        if evaluate_constraints:
            result = model.evaluate_constraints(trajectory)
        else:
            result = None
        return result

    def simulate_for_initialisation(self, model: Model, horizon: float, initialization_constraints: Set[Observable]) \
            -> (bool, float, Dict[str, List[float]]):
        trajectory: Dict[str, List[float]] = self.simulate_trajectory(model, horizon)
        trajectory_length = len(trajectory['time'])
        init_time_idx = 0
        for obs in initialization_constraints:
            # init_value = next((v for v in trajectory[obs.id] if v <= 0), None)
            obs_init_time_idx: int = model.compute_first_satisfaction_index(obs, trajectory)
            if obs_init_time_idx == trajectory_length:
                return False, trajectory_length, dict()
            init_time_idx = max(init_time_idx, obs_init_time_idx)

        trajectory_up_to_init = dict()
        for obs_id, obs_traj in trajectory.items():
            # returns the trajectory up to the moment when the model is initialized
            trajectory_up_to_init[obs_id] = obs_traj[:init_time_idx]

        return True, trajectory['time'][init_time_idx], trajectory_up_to_init

    @abstractmethod
    def load_model(self, model_filename: str, **kwargs) -> ModelInstance:
        pass

    @abstractmethod
    def simulate_trajectory(self, model: Model, horizon: float) -> Dict[str, List[float]]:
        pass

    @abstractmethod
    def simulate_trajectory_and_set(self, model: Model, horizon: float, exclude=None) -> Dict[str, List[float]]:
        pass

    @abstractmethod
    def simulate_trajectory_and_get_state(self, model: Model, horizon: float, exclude=None) -> Dict[str, float]:
        pass

    @abstractmethod
    def restore_state(self, model: Model, changed_values: dict) -> None:
        pass

    @abstractmethod
    def model_instance_class(self) -> Type[ModelInstance]:
        pass



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020 Marco Esposito, Leonardo Picchiami.
"""

from unittest import TestCase, main

from apricopt.model.Model import Model
from apricopt.model.Observable import Observable
from apricopt.model.Parameter import Parameter
from apricopt.simulation.MockUp.MockUpSimulationEngine import MockUpSimulationEngine
from apricopt.solving.blackbox.MockUp.MockUpSolver import MockUpSolver
from apricopt.tests.solving.blackbox.mockup.MockUpBlackBox import MockUpBlackBox


class TestMockUp(TestCase):
    def test(self):
        sim_engine = MockUpSimulationEngine()
        black_box: MockUpBlackBox = MockUpBlackBox(sim_engine)
        solver = MockUpSolver()
        iterations = 1000
        min_params, min_obj, _, _, _ = solver.solve(black_box, [iterations])
        print(f"\n\n==================================\nSolution: {min_params}\nBest Objective Value: {min_obj}")


if __name__ == "__main__":
    main()
